import abc
import datetime
import errno
import getpass
import os
import shutil
import tempfile
import weakref
from collections import defaultdict
from typing import Dict, List
from .types import default

import coolname


WEBAPI_TEST_MODE_SHARED_CONFIG = os.getenv('WEBAPI_TEST_MODE_SHARED_CONFIG')
if os.getenv('OTP_WEBAPI_TEST_MODE') and not WEBAPI_TEST_MODE_SHARED_CONFIG:
    raise ValueError("WEBAPI_TEST_MODE_SHARED_CONFIG is not set, but it is required with OTP_WEBAPI_TEST_MODE.")


def default_clean_up(clean_up):
    from ..configuration import config
    if clean_up is default:
        return config.clean_up_tmp_files
    return clean_up


def get_logger(*args):
    from .. import log
    return log.get_logger(__name__, *args)


class CleanUpFinalizer:
    """
    Class that manages destruction of the object,
    setting up proper finalizer callback and clean_up boolean flag.
    """

    def __init__(self, clean_up=default, *args):
        clean_up = default_clean_up(clean_up)
        # we need to use reference type here so finalizer can get the latest clean_up value,
        # even if this value is changed after finalizer is created
        # (because weakref.finalize can't use object's fields)
        self._clean_up_ref = [clean_up]
        self._finalizer = weakref.finalize(self, self._cleanup, self._clean_up_ref, *args)

    @classmethod
    def _cleanup(cls, clean_up_ref, *args):
        raise NotImplementedError()

    @property
    def need_to_cleanup(self):
        return self._clean_up_ref[0]

    @need_to_cleanup.setter
    def need_to_cleanup(self, value):
        self._clean_up_ref[0] = value


class File(abc.ABC):
    def __init__(self, path):
        self._path = path

    @property
    def path(self):
        return self._path

    def __str__(self):
        return self._path

    def __fspath__(self):
        """
        If children inherit os.PathLike, then this method is required and
        makes files behave compatible with 'os' module
        """
        return self._path


def __name_generator():
    """
    Returns tuple with initialized name generator and
    the number of unique names this generator will produce
    before it starts to repeat itself.
    """
    def cool_gen():
        while True:
            yield coolname.generate_slug(2)
    return cool_gen(), coolname.get_combinations_count(2)


def mktemp(fun, dir, prefix='', suffix='', **kwargs):
    names, max_unique_values = __name_generator()
    for _ in range(max_unique_values):
        name = next(names)
        filename = os.path.join(dir, prefix + name + suffix)
        try:
            return fun(filename, **kwargs), filename
        except FileExistsError:
            continue
        except PermissionError:
            # This exception is thrown when a directory with the chosen name already exists on Windows
            if os.name == 'nt' and os.path.isdir(dir) and os.access(dir, os.W_OK):
                continue
            else:
                raise
    raise FileExistsError(errno.EEXIST, 'No usable temporary file name found')


def mkstemp(dir, prefix='', suffix=''):
    flags = os.O_RDWR | os.O_CREAT | os.O_EXCL
    if hasattr(os, 'O_NOFOLLOW'):
        flags |= os.O_NOFOLLOW
    return mktemp(os.open, dir, prefix, suffix, flags=flags, mode=0o600)


def mkdtemp(dir, prefix='', suffix=''):
    _, filename = mktemp(os.mkdir, dir, prefix, suffix, mode=0o700)
    return filename


_TMP_CONFIGS_DIR_BASE = os.environ.get('OTP_BASE_FOLDER_FOR_GENERATED_RESOURCE',
                                       os.path.join(tempfile.gettempdir(), "test_" + getpass.getuser()))


class TmpFile(File, os.PathLike, CleanUpFinalizer):

    ALL: Dict[str, List['TmpFile']] = defaultdict(list)
    keep_everything_generated = False

    def __init__(self, suffix="", name="", clean_up=default, force=False, base_dir=None):
        """
        Class to create a temporary file.
        By default, this file will be deleted automatically after all references to it are gone.
        Base path where temporary files are created could be set using the ``ONE_TICK_TMP_DIR``.
        By default they are created under the ``tempfile.gettempdir()`` folder.


        Parameters
        ----------
        name: str
            name of the temporary file without suffix.
            By default some random name will be generated.
        suffix: str
            suffix of the name of the temporary file.
        clean_up: bool
            Controls whether this temporary file will be deleted automatically
            after all references to it are gone.

            By default,
            :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.
        force: bool
            Rewrite temporary file if it exists and parameter ``name`` is set.
        base_dir: str
            Absolute path of the directory where temporary file will be created.

        See also
        --------

        The testing framework has a ``--keep-generated`` flag that controls clean up for all related instances
        :ref:`onetick py test features`
        """
        clean_up = default_clean_up(clean_up)
        clean_up = clean_up and not TmpFile.keep_everything_generated
        fd, self._path = self._create(clean_up, suffix=suffix, name=name, force=force, base_dir=base_dir)
        # we only needed to create file, so closing opened file descriptor
        os.close(fd)
        self._logger = get_logger(self.__class__.__name__)
        CleanUpFinalizer.__init__(self, clean_up, self.path, self._logger)
        self._logger.debug(f'created {self._path}, clean_up={self.need_to_cleanup}')

    def __repr__(self):
        return f"TmpFile({self.path})"

    @classmethod
    def _cleanup(cls, clean_up_ref, path, logger):
        if clean_up_ref[0] and os.path.exists(path):
            logger.debug(f'removing {path}')
            try:
                os.remove(path)
            except Exception:
                # TODO: remove try-except block when BDS-116 will be fixed
                pass

    def _create(self, clean_up, suffix="", name="", force=False, base_dir=None):

        self._parent_dir = None

        if ONE_TICK_TMP_DIR() and not base_dir:
            # allows to set a test name suffix for when we use it in
            # tests with goal to distinguish tests content (configs, databases, etc)
            # from each other
            dir_path = ONE_TICK_TMP_DIR()
            # save to access them from the teardown
            TmpFile.ALL[dir_path].append(self)

            if not os.path.exists(dir_path):
                # If it does not exist, then lets create it. Note that it goes up
                # recursively until finds an existing folder.
                # We store it as a field of the instance to guarantee that it will
                # be destroyed later than a temporary file.
                self._parent_dir = GeneratedDir(dir_path, clean_up=clean_up)
            elif dir_path in GeneratedDir.ALL:
                self._parent_dir = GeneratedDir.ALL[dir_path]
        else:
            if base_dir and not os.path.exists(base_dir):
                # If it does not exist, then lets create it. Note that it goes up
                # recursively until finds an existing folder.
                # We store it as a field of the instance to guarantee that it will
                # be destroyed later than a temporary file.
                self._parent_dir = GeneratedDir(base_dir, clean_up=clean_up)
            dir_path = base_dir if base_dir else TMP_CONFIGS_DIR()
            TmpFile.ALL[dir_path].append(self)
            if dir_path in GeneratedDir.ALL:
                self._parent_dir = GeneratedDir.ALL[dir_path]

        # let know the parent generated folder to not destroy itself
        if self._parent_dir is not None and not clean_up:
            self._parent_dir.need_to_cleanup = False

        if name:
            path = os.path.normpath(os.path.join(dir_path, name)) + suffix
            flags = os.O_RDWR | os.O_CREAT
            if not force:
                flags = flags | os.O_EXCL
            if hasattr(os, 'O_NOFOLLOW'):
                flags |= os.O_NOFOLLOW
            mode = 0o600
            if os.getenv('OTP_WEBAPI_TEST_MODE'):
                mode = 0o644
            return os.open(path, flags=flags, mode=mode), path
        else:
            return mkstemp(dir=dir_path, suffix=suffix, prefix="")


class TmpDir(os.PathLike, CleanUpFinalizer):

    ALL: Dict[str, List['TmpDir']] = defaultdict(list)
    keep_everything_generated = False

    def __init__(self, rel_path="", *, suffix="", clean_up=default, base_dir=""):
        """
        Class to create a temporary directory.
        By default, this directory will be deleted automatically after all references to it are gone.
        All files and directories under this one will be deleted too.

        Base path where directories are created could be set using the ``ONE_TICK_TMP_DIR``.
        By default they are created under the ``tempfile.gettempdir()`` folder.

        Parameters
        ----------
        rel_path: str
            relative path to the temporary directory.
            If empty, then the name will be auto-generated.
        suffix: str
            suffix of the name of the temporary directory.
        base_dir: str
            relative path of the directory where temporary directory will be created.
        clean_up: bool
            Controls whether this temporary directory will be deleted automatically
            after all references to it are gone.

            By default,
            :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.

        See also
        --------

        The testing framework has a ``--keep-generated`` flag that controls clean up for all related instances
        :ref:`onetick py test features`
        """
        clean_up = default_clean_up(clean_up)
        clean_up = clean_up and not TmpDir.keep_everything_generated
        self._parent_dir = None

        if os.path.isabs(rel_path):
            raise ValueError("Absolute paths are not supported in 'rel_path' parameter.")

        if ONE_TICK_TMP_DIR():
            # allows to set a test name suffix for when we use it in
            # tests with goal to distinguish tests content (configs, databases, etc)
            # from each other
            dir_path = os.path.normpath(os.path.join(ONE_TICK_TMP_DIR(), base_dir))

            if not os.path.exists(dir_path):
                # If it does not exist, then lets create it. Note that it goes up
                # recursively until finds an existing folder.
                # We store it as a field of the instance to guarantee that it will
                # be destroyed later than a temporary file.
                self._parent_dir = GeneratedDir(dir_path, clean_up=clean_up)
            elif not clean_up:
                if dir_path in GeneratedDir.ALL:  # NOSONAR
                    # let know the parent generated folder to not destroy itself
                    GeneratedDir.ALL[dir_path].need_to_cleanup = False

                    self._parent_dir = GeneratedDir.ALL[dir_path]

        else:
            dir_path = os.path.normpath(os.path.join(TMP_CONFIGS_DIR(), base_dir))

        if rel_path:
            path = os.path.normpath(os.path.join(dir_path, rel_path)) + suffix
            # dir_path should be the parent directory of path
            dir_path = os.path.dirname(path)
            for tmp_dir in TmpDir.ALL[dir_path]:
                # check if path already exists
                if tmp_dir.path == path:
                    self.path = tmp_dir.path
                    if not clean_up:
                        tmp_dir.need_to_cleanup = clean_up
                    return

            os.mkdir(path, mode=0o700)
            self.path = path
        else:
            self.path = mkdtemp(dir=dir_path, suffix=suffix, prefix="")

        # save to access them in the teardown
        TmpDir.ALL[dir_path].append(self)

        self._logger = get_logger(self.__class__.__name__)
        CleanUpFinalizer.__init__(self, clean_up, self.path, self._logger)
        self._logger.debug(f'created {self.path}, clean_up={self.need_to_cleanup}')

    @CleanUpFinalizer.need_to_cleanup.setter   # type: ignore
    def need_to_cleanup(self, value):
        super(TmpDir, type(self)).need_to_cleanup.fset(self, value)
        children = self.ALL[self.path]
        for child in children:
            child.need_to_cleanup = value

    @classmethod
    def _cleanup(cls, clean_up_ref, path, logger):
        if clean_up_ref[0]:
            logger.debug(f'removing {path}')
            shutil.rmtree(path, ignore_errors=True)

    def __str__(self):
        return self.path

    def __repr__(self):
        return f"TmpDir({self.path})"

    def __fspath__(self):
        return self.path


class GeneratedDir(os.PathLike, CleanUpFinalizer):

    ALL: Dict[str, 'GeneratedDir'] = {}  # store all created dir with goal to set cleanup=False if it requires
    keep_everything_generated = False

    def __init__(self, dir_path, clean_up=default):
        """
        Class to create generated temporary directory.
        By default, this directory will be deleted automatically after all references to it are gone.
        All files and directories under this one will be deleted too.

        The main difference from TmpDir class is that also this directory's parent directory will be deleted!

        The exact path to this directory will be: ``dir_path``.

        Parameters
        ----------
        dir_path: str
            path to the temporary directory.
        clean_up: bool
            Controls whether this temporary directory will be deleted automatically
            after all references to it are gone.

            By default,
            :py:attr:`otp.config.clean_up_tmp_files<onetick.py.configuration.Config.clean_up_tmp_files>` is used.
        """

        parent = os.path.normpath(os.path.dirname(dir_path))

        clean_up = default_clean_up(clean_up)
        clean_up = clean_up and not GeneratedDir.keep_everything_generated

        self._parent_dir = None
        if not os.path.exists(parent):
            # Do the same thing for the parent folder.
            # Save it as a field to guarantee that child folder is destroyed
            # earlier than parent
            self._parent_dir = GeneratedDir(parent, clean_up=clean_up)
        elif not clean_up:
            if parent in GeneratedDir.ALL:  # NOSONAR
                GeneratedDir.ALL[parent].need_to_cleanup = False

        if self._parent_dir is None and parent in GeneratedDir.ALL:
            self._parent_dir = GeneratedDir.ALL[parent]

        self.path = dir_path

        if os.path.exists(dir_path):
            # Do not create and even handle it, because it has been created externally.
            # It mgiht happen in the concurrent runs.
            return

        # We need to set exist_ok=True, because it still can be created externally in the concurrent runs
        os.makedirs(dir_path, exist_ok=True)

        self._logger = get_logger(self.__class__.__name__)
        CleanUpFinalizer.__init__(self, clean_up, self.path, self._logger)
        self._logger.debug(f'created {self.path}, clean_up={self.need_to_cleanup}')

        GeneratedDir.ALL[dir_path] = self

    @CleanUpFinalizer.need_to_cleanup.setter   # type: ignore
    def need_to_cleanup(self, value):
        super(GeneratedDir, type(self)).need_to_cleanup.fset(self, value)

        if self._parent_dir:
            self._parent_dir.need_to_cleanup = value

    @classmethod
    def _cleanup(cls, clean_up_ref, path, logger):
        if clean_up_ref[0]:
            logger.debug(f'removing {path}')
            shutil.rmtree(path, ignore_errors=True)

    def __str__(self):
        return self.path

    def __repr__(self):
        return f"GeneratedDir({self.path})"

    def __fspath__(self):
        return self.path


# ----------------------------------------------------
# We set TMP_CONFIGS_DIR to a function, that returns
# only the first result.
# We need to support multiprocessing environment, where
# every process should use it's own base temporary folder
# but generate it only when something is required, because
# otherwise parent process could construct it with its pid
# and other processes would use the same location.


def _gen_root_path_():
    generated = []

    def _inner_():
        if not generated:
            generated.append(
                GeneratedDir(
                    os.path.join(
                        _TMP_CONFIGS_DIR_BASE,
                        f'run_{datetime.datetime.now().strftime("%Y%m%d_%H%M%S")}_{os.getpid()}',
                    )
                ).path
            )
        return generated[0]

    return _inner_


TMP_CONFIGS_DIR = _gen_root_path_()
def ONE_TICK_TMP_DIR():  # noqa # NOSONAR
    if os.getenv('OTP_WEBAPI_TEST_MODE'):
        return WEBAPI_TEST_MODE_SHARED_CONFIG
    if 'ONE_TICK_TMP_DIR' not in os.environ:
        return None
    return os.path.normpath(os.path.join(TMP_CONFIGS_DIR(), os.environ['ONE_TICK_TMP_DIR']))
# ----------------------------------------------------


class PermanentFile(File):
    def copy(self):
        return TmpFile(self._path)



from typing import List, Union

import onetick.py as otp
from onetick.py.backports import Literal
from onetick.py.core.source import Source
from onetick.py.otq import otq

from .. import types as ott
from .. import utils
from ..compatibility import is_supported_point_in_time
from .common import update_node_tick_type


def PointInTime(  # NOSONAR
    source: 'Source',
    times: List[Union[str, ott.datetime]],
    offsets: List[int],
    offset_type: Literal['time_msec', 'num_ticks'] = 'time_msec',
    db=utils.adaptive_to_default,
    tick_type=utils.adaptive,
    symbol=utils.adaptive_to_default,
    start=utils.adaptive,
    end=utils.adaptive,
) -> 'Source':
    """
    This function propagates ticks from ``source`` that are offset by
    the specified number of milliseconds or by the specified number of ticks
    relative to the timestamps specified in ``times``.

    Output tick may be generated for each specified timestamp and offset pair.

    If ``source`` doesn't have a tick with specified timestamp and offset, then output tick is not generated.

    Fields **TICK_TIME** and **OFFSET** are also added to the output ticks,
    specifying original timestamp of the tick and the offset that was specified for it.

    Note
    ----
    In order for this method to have reasonable performance,
    the set of queried timestamps has to be relatively small.

    In other words, the points in time, which the user is interested in,
    have to be quite few in order usage of this method to be justified.

    Parameters
    ----------
    source:
        The source from which the data will be queried.
    times:
        List of timestamps to get query ticks from.
    offsets:
        List of integers specifying offsets for each timestamp.
    offset_type: 'time_msec' or 'num_ticks'
        The type of offset: number of milliseconds or the number of ticks.
    db: str
        This parameter is used to set database part of a tick type of the OneTick graph node.
        Should not be specified in most cases,
        default value will be deduced depending on configuration and query structure.
    tick_type: str
        This parameter is used to set tick type part of a tick type of the OneTick graph node.
        Should not be specified in most cases,
        default value will be deduced depending on configuration and query structure.
    symbol: str
        This parameter is used to set bound symbol of the OneTick graph node.
        Should not be specified in most cases,
        default value will be deduced depending on configuration and query structure.
    start:
        Can be used to specify custom start time for this source.
        By default start time of the query will be inherited when running the query.
    end:
        Can be used to specify custom start time for this source.
        By default end time of the query will be inherited when running the query.

    See also
    --------
    | **POINT_IN_TIME** OneTick event processor
    | :meth:`onetick.py.Source.point_in_time`
    | :func:`onetick.py.join_by_time`

    Examples
    --------

    Quotes for testing:

    .. testcode::

       qte = otp.Ticks(ASK_PRICE=[20, 21, 22, 23, 24, 25], BID_PRICE=[20, 21, 22, 23, 24, 25])
       print(otp.run(qte))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE
       0 2003-12-01 00:00:00.000         20         20
       1 2003-12-01 00:00:00.001         21         21
       2 2003-12-01 00:00:00.002         22         22
       3 2003-12-01 00:00:00.003         23         23
       4 2003-12-01 00:00:00.004         24         24
       5 2003-12-01 00:00:00.005         25         25

    Getting quotes exactly at specified timestamps:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = otp.PointInTime(qte,
                              times=[otp.dt(2003, 12, 1, 0, 0, 0, 1000), '20031201000000.003'],
                              offsets=[0])
       print(otp.run(data))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE               TICK_TIME  OFFSET
       0 2003-12-01 00:00:00.001         21         21 2003-12-01 00:00:00.001       0
       1 2003-12-01 00:00:00.003         23         23 2003-12-01 00:00:00.003       0


    Offset may be positive or negative.
    If several offsets are specified, several output ticks may be generated for a single timestamp:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = otp.PointInTime(qte,
                              times=[otp.dt(2003, 12, 1, 0, 0, 0, 3000)],
                              offsets=[0, 1])
       print(otp.run(data))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE               TICK_TIME  OFFSET
       0 2003-12-01 00:00:00.003         23         23 2003-12-01 00:00:00.003       0
       1 2003-12-01 00:00:00.003         24         24 2003-12-01 00:00:00.004       1

    By default the number of milliseconds is used as an offset.
    You can also specify the number of ticks as an offset:

    .. testcode::
       :skipif: not is_supported_point_in_time()

       data = otp.PointInTime(qte,
                              times=[otp.dt(2003, 12, 1, 0, 0, 0, 3000)],
                              offsets=[-1, 1],
                              offset_type='num_ticks')
       print(otp.run(data))

    .. testoutput::

                            Time  ASK_PRICE  BID_PRICE               TICK_TIME  OFFSET
       0 2003-12-01 00:00:00.003         22         22 2003-12-01 00:00:00.002      -1
       1 2003-12-01 00:00:00.003         24         24 2003-12-01 00:00:00.004       1
    """
    if not is_supported_point_in_time():
        raise RuntimeError('PointInTime event processor is not supported on this OneTick version')

    res = otp.Source(_symbols=symbol, _start=start, _end=end)

    times = [
        t if isinstance(t, str) else ott._format_datetime(t, '%Y%m%d%H%M%S.%f', add_nano_suffix=True)
        for t in times or []
    ]

    if offset_type not in ('time_msec', 'num_ticks'):
        raise ValueError(f"Wrong value for parameter 'offset_type': {offset_type}")

    query_name = source._store_in_tmp_otq(
        res._tmp_otq,
        operation_suffix='point_in_time',
        # set default symbol, even if it's not set by user, symbol's value doesn't matter in this case
        symbols=otp.config.get('default_symbol', 'ANY')
    )
    otq_query = f'THIS::{query_name}'

    pit_params = dict(
        otq_query=otq_query,
        offset_type=offset_type.upper(),
        offsets=','.join(map(str, offsets)),
        times=','.join(map(str, times)),
    )
    res.source(otq.PointInTime(**pit_params))

    res.schema.set(**source.schema)
    res.schema.update(
        **{
            'TICK_TIME': otp.nsectime,
            'OFFSET': int,
        }
    )

    update_node_tick_type(res, tick_type, db)

    return res



import os
from datetime import datetime
from typing import Iterable, Type, Union, Optional
from contextlib import suppress, contextmanager
from textwrap import dedent

import dotenv
from onetick.py.otq import otq
from .utils import default_license_dir, default_license_file, get_local_number_of_cores
import onetick.py.types as ott

DEFAULT_LICENSE_DIR = default_license_dir()
DEFAULT_LICENSE_FILE = default_license_file()

DATETIME_FORMATS = (
    '%Y/%m/%d %H:%M:%S.%f',
    '%Y/%m/%d %H:%M:%S',
)
DATETIME_DESCRIPTION = (
    "Format of the env variable: "
    f"{', '.join(':code:`{}`'.format(fmt) for fmt in DATETIME_FORMATS)}."
)


def parse_datetime(s):
    for fmt in DATETIME_FORMATS:
        with suppress(ValueError):
            return datetime.strptime(s, fmt)
    raise ValueError(
        f"The datetime pattern is not supported for string '{s}'. "
        f"Available patterns: {DATETIME_FORMATS}"
    )


def _env_func_concurrency(value: str) -> Optional[int]:
    if not value:
        return None
    if value == 'local_number_of_cores':
        return get_local_number_of_cores()
    return int(value)


def default_query_concurrency():
    concurrency = config.default_concurrency
    if concurrency is None:
        from onetick.py.compatibility import is_zero_concurrency_supported
        # TODO: this logic should be in the default value of otp.config.default_concurrency,
        # but there are complex problems with circular import in this case
        if is_zero_concurrency_supported():
            concurrency = 0
        else:
            concurrency = 1
    return concurrency


def default_presort_concurrency():
    if config.presort_force_default_concurrency:
        return default_query_concurrency()
    else:
        return None


class _nothing(type):
    def __repr__(cls):
        return cls.__name__


class nothing(metaclass=_nothing):
    """
    This is nothing.
    """


class OtpProperty:
    """
    .. attribute:: {name}
       :type: {base_type}
       :value: {base_value}

       {description}

       {env_var_name}

       {env_var_desc}
    """
    def __init__(self, description, base_default, env_var_name=None, env_var_func=None,
                 env_var_desc=None, set_value=nothing, allowed_types: Union[Type, Iterable] = nothing,
                 validator_func=None):
        self._base_default = base_default
        self._env_var_name = env_var_name
        self._env_var_func = env_var_func
        self._env_var_desc = env_var_desc
        self._set_value = set_value
        self._description = description
        if self._base_default is nothing:
            self._allowed_types = []
        else:
            self._allowed_types = [type(self._base_default)]
        if allowed_types is not nothing:
            if isinstance(allowed_types, Iterable):
                self._allowed_types.extend(allowed_types)
            else:
                self._allowed_types.append(allowed_types)
        self._allowed_types = tuple(set(self._allowed_types))  # type: ignore[assignment]
        # will be monkeypatched later
        self._name = None
        if validator_func is None:
            self._validator_func = lambda x: x
        else:
            self._validator_func = validator_func

    def _get_doc(self, name):
        env_var_name = ''
        env_var_desc = ''
        if self._env_var_name is not None:
            env_var_name = f'Can be set using environment variable :envvar:`{self._env_var_name}`.'
            if self._env_var_desc is not None:
                env_var_desc = self._env_var_desc
        return self.__class__.__doc__.format(
            name=name,
            description=self._description,
            base_type=','.join(t.__name__ for t in self._allowed_types),
            base_value=repr(self._base_default),
            env_var_name=env_var_name,
            env_var_desc=env_var_desc
        )

    def __get__(self, obj, objtype=None):
        if self._set_value is not nothing:
            return self._set_value
        if self._env_var_name:
            env_var_value = os.environ.get(self._env_var_name, None)
            if env_var_value is not None:
                if self._env_var_func:
                    return self._env_var_func(env_var_value)
                return env_var_value
        if obj is not None:
            # get value from default config
            if self._env_var_name and self._env_var_name in obj.default_config:
                var_value = obj.default_config[self._env_var_name]
                if self._env_var_func:
                    return self._env_var_func(var_value)
                return var_value
        if self._base_default is nothing:
            raise ValueError(f'onetick.py.config.{self._name} is not set!')
        return self._base_default

    def __set__(self, obj, value):
        # assigning to nothing is permitted
        # assigning to nothing will reset value to default
        if not isinstance(value, self._allowed_types) and value is not nothing:
            raise ValueError(f'Type of passed configuration value "{type(value)}" should be one of '
                             f'the allowed types for this configuration {self._allowed_types}')
        if value is nothing:
            self._set_value = value
        else:
            self._set_value = self._validator_func(value)


class OtpDerivedProperty:
    """
    .. attribute:: {name}
       :type: {base_type}
       :value: {base_value}

       {description}
    """

    def __init__(self, description, definition_function):
        self._description = description
        self.__definition_function = definition_function

    def __get__(self, obj, objtype=None):
        return self.__definition_function(obj)

    def _get_doc(self, name, base_object):
        value = self.__definition_function(base_object, docs=True)
        return self.__doc__.format(
            name=name,
            description=self._description,
            base_type=type(value).__name__,
            base_value=value,
        )

    def __set__(self, obj, value):
        raise AttributeError('It\'s not allowed to change a derived property. Change source properties, and its value '
                             'will be updated automatically.')


class OtpShowStackInfoProperty(OtpProperty):
    """
    .. attribute:: {name}
       :type: {base_type}
       :value: {base_value}

       {description}
    """
    @staticmethod
    def parser(value):
        return str(value).lower() in ('1', 'true', 'yes')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # set default value on module loading
        self.__set_in_onetick_query__()

    def __get__(self, obj, objtype=None):
        value = super().__get__(obj, objtype)
        return self.parser(value)

    def __set__(self, obj, value):
        super().__set__(obj, value)
        self.__set_in_onetick_query__()

    def __set_in_onetick_query__(self):
        value = 1 if self.__get__(None) else 0
        otq.API_CONFIG['SHOW_STACK_INFO'] = value


def document_config(cls):
    manually_set_properties_doc = ''
    changeable_properties_doc = ''
    derived_properties_doc = ''

    manually_set_options = cls.manually_set_options()
    for b in manually_set_options:
        manually_set_properties_doc += cls.__dict__[b]._get_doc(b)
    for c in cls.get_changeable_config_options():
        cls.__dict__[c]._name = c
        if c not in manually_set_options:
            changeable_properties_doc += cls.__dict__[c]._get_doc(c)
    for d in cls.get_derived_config_options():
        cls.__dict__[d]._name = d
        derived_properties_doc += cls.__dict__[d]._get_doc(d, base_object=cls)

    cls.__doc__ = cls.__doc__.format(
        manually_set_properties=manually_set_properties_doc,
        changeable_properties=changeable_properties_doc,
        derived_properties=derived_properties_doc,
    )
    cls.__doc__ = dedent(cls.__doc__)

    return cls


@document_config
class Config:
    """
    This object is used to access ``onetick.py`` configuration variables.

    Configuration variables may be accessed via :code:`otp.config['...']` syntax, e.g. :code:`otp.config['tz']`.

    Configuration variables may be changed by:

    * during python runtime by modifying properties of object ``otp.config``,
    * by setting environment variables *before* importing ``onetick.py`` module.

    During python runtime you can modify properties of ``otp.config`` object either directly or via context manager:
    ``with otp.config('property', 'value'):``

    Also special environment variable ``OTP_DEFAULT_CONFIG_PATH`` can be used to specify a file,
    from which configuration variables will be taken.
    This file will be read only once on module loading or when getting one of the configuration variables
    when the environment variable is discovered.
    The names of the variables in this file are the same as the names of environment variables.

    In case several methods of setting configuration variables are used,
    the following order of priority is in place:

    1. Value that is set by modifying object ``otp.config``
    2. Value that is set via environment variable
    3. Value that is set in file ``OTP_DEFAULT_CONFIG_PATH``
    4. Default value specified in the source code

    To reset configuration value that has been set by modifying object ``otp.config``,
    special value ``otp.config.default`` should be assigned to it.

    Most of the config vars are optional and have default values,
    but some of them need to be set manually.

    There are also some environment variables that do not have
    corresponding property in ``otp.config`` object:

    * ``OTP_BASE_FOLDER_FOR_GENERATED_RESOURCE``:
      a folder where all intermediate queries, files and databases
      generated by ``onetick-py`` are located.
      The default value is system-dependent, e.g. some generated
      directory with a unique name under a standard directory **/tmp** for Linux.

    **The following properties must be set manually in most cases:**
    {manually_set_properties}

    **The following properties can be changed:**
    {changeable_properties}

    **The following properties are derived and thus read-only:**
    {derived_properties}
    """
    __default_config = None

    @property
    def default_config(self):
        default_config_path = os.environ.get('OTP_DEFAULT_CONFIG_PATH')
        if not default_config_path:
            return {}
        if self.__default_config is None:
            default_config = dotenv.dotenv_values(default_config_path)
            available_option_names = []
            for name, option in self.get_changeable_config_options().items():
                if option._env_var_name:
                    available_option_names.append(option._env_var_name)
            diff = set(default_config).difference(available_option_names)
            if diff:
                raise ValueError(f'Configuration options {diff} from file'
                                 f' OTP_DEFAULT_CONFIG_PATH="{default_config_path}" are not supported.'
                                 f' Available options: {available_option_names}.')
            Config.__default_config = default_config
        return self.__default_config or {}

    def __getitem__(self, item):
        if item not in self.__class__.__dict__.keys():
            raise AttributeError(f'"{item}" is not in the list of onetick.py config options!')
        return self.__class__.__dict__[item].__get__(self)

    def __setitem__(self, item, value):
        if item not in self.__class__.__dict__.keys():
            raise AttributeError(f'"{item}" is not in the list of onetick.py config options!')
        self.__class__.__dict__[item].__set__(self, value)

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except ValueError:
            return default

    def __setattr__(self, item, value):
        """
        To avoid accidental declaration of non-existing properties, e.g. `otp.config.timezone = "GMT"`
        """
        self.__setitem__(item, value)

    @contextmanager
    def __call__(self, name, value):
        old_value = self[name]
        try:
            self[name] = value
            yield
        finally:
            self[name] = old_value

    @classmethod
    def get_changeable_config_options(cls):
        """
        useful for tests where you may want to memorize all existing configuration options before changing them
        """
        return {
            option: value
            for option, value in cls.__dict__.items()
            if isinstance(value, OtpProperty)
        }

    @classmethod
    def get_derived_config_options(cls):
        return {
            option: value
            for option, value in cls.__dict__.items()
            if isinstance(value, OtpDerivedProperty)
        }

    @classmethod
    def manually_set_options(cls):
        return {
            option: value
            for option, value in cls.__dict__.items()
            if isinstance(value, OtpProperty) and value._base_default is nothing
        }

    default = nothing

    tz = OtpProperty(
        description='Default timezone used for running queries and creating databases, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`. '
                    'Default value is the local timezone of your machine.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_DEFAULT_TZ',
    )

    context = OtpProperty(
        description='Default context used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`.'
                    'Note: In WebAPI mode it will have `None` value.',
        base_default='DEFAULT' if not otq.webapi else None,
        allowed_types=[str],
        env_var_name='OTP_CONTEXT',
    )

    default_start_time = OtpProperty(
        description='Default start time used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`.',
        base_default=nothing,
        allowed_types=[datetime, ott.datetime],
        env_var_name='OTP_DEFAULT_START_TIME',
        env_var_func=parse_datetime,
        env_var_desc=DATETIME_DESCRIPTION,
    )

    default_end_time = OtpProperty(
        description='Default end time used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`.',
        base_default=nothing,
        allowed_types=[datetime, ott.datetime],
        env_var_name='OTP_DEFAULT_END_TIME',
        env_var_func=parse_datetime,
        env_var_desc=DATETIME_DESCRIPTION,
    )

    default_db = OtpProperty(
        description='Default database name used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`.',
        base_default=nothing,
        allowed_types=str,
        env_var_name='OTP_DEFAULT_DB',
    )

    default_symbol = OtpProperty(
        description='Default symbol name used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`.',
        base_default=nothing,
        allowed_types=str,
        env_var_name='OTP_DEFAULT_SYMBOL',
    )

    default_symbology = OtpProperty(
        description='Default database symbology.',
        base_default='BZX',
        env_var_name='OTP_DEFAULT_SYMBOLOGY',
    )

    def _default_db_symbol(obj, docs=False):  # noqa
        try:
            return obj.default_db + '::' + obj.default_symbol
        except (ValueError, TypeError):
            if not docs:
                raise

    default_db_symbol = OtpDerivedProperty(
        description='Default symbol with database. '
                    'Defined with :py:attr:`default_db` and :py:attr:`default_symbol` '
                    'as string **default_db::default_symbol**.',
        definition_function=_default_db_symbol,
    )

    def _default_date(obj, docs=False):  # noqa
        try:
            return datetime.combine(obj.default_start_time.date(), datetime.min.time())
        except (ValueError, TypeError, AttributeError):
            if not docs:
                raise

    default_date = OtpDerivedProperty(
        description='Default date. '
                    'Defined as a date part of :py:attr:`default_start_time`.',
        definition_function=_default_date,
    )

    default_concurrency = OtpProperty(
        description='Default concurrency level used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`. '
                    'Default value is ``None`` which means that the value is adaptive '
                    'and is set to 0 (meaning concurrency will be auto-assigned by OneTick server) '
                    'on the latest OneTick versions where it is supported '
                    'or to 1 (meaning no concurrency) on older versions. '
                    'Special value ``local_number_of_cores`` can be used to set concurrency '
                    'to the number of cores of the machine where python code executes '
                    '(this corresponds to the previous default logic still expected by some users).',
        base_default=None,
        allowed_types=[type(None), int, str],
        env_var_name='OTP_DEFAULT_CONCURRENCY',
        env_var_func=_env_func_concurrency,
        validator_func=lambda x: _env_func_concurrency(str(x) if x is not None else '')
    )

    presort_force_default_concurrency = OtpProperty(
        description='By default concurrency value for PRESORT EPs is empty '
                    'and inherited from the concurrency level set in the query where this EP is used. '
                    'However, is some cases it may be desirable '
                    'to force setting default concurrency level for all PRESORT EPs '
                    '(this corresponds to the previous default logic still expected by some users), '
                    'for example when PRESORT is located in the first stage query and cannot inherit '
                    'concurrency from the main query.',
        base_default=False,
        allowed_types=[bool],
        env_var_name='OTP_PRESORT_FORCE_DEFAULT_CONCURRENCY',
        env_var_func=OtpShowStackInfoProperty.parser,
    )

    # default batch size is set to 0, so the number of symbols in batch is not limited
    # it should work better in simple cases, but may use too much memory for complex queries
    default_batch_size = OtpProperty(
        description='Default batch size used for running queries, '
                    'e.g. with :py:func:`otp.run<onetick.py.run>`. '
                    'Batch size is the maximum number of symbols that are processed at once. '
                    'The value of 0 means unlimited -- works faster for simple queries, '
                    'but may consume too much memory for complex queries.',
        base_default=0,
        env_var_name='OTP_DEFAULT_BATCH_SIZE',
        env_var_func=int,
    )

    default_license_dir = OtpProperty(
        description='Default path for license directory. '
                    'Needed for user to be allowed to use OneTick API. '
                    'Default value is system-dependent: '
                    '**/license** for Linux systems and '
                    '**C:/OMD/client_data/config/license_repository** for Windows systems.',
        base_default=DEFAULT_LICENSE_DIR,
        env_var_name='OTP_DEFAULT_LICENSE_DIR',
        allowed_types=str,
    )

    default_license_file = OtpProperty(
        description='Default path for license file. '
                    'Needed for user to be allowed to use OneTick API. '
                    'Default value is system-dependent: '
                    '**/license/license.dat** for Linux systems and '
                    '**C:/OMD/client_data/config/license.dat** for Windows systems.',
        base_default=DEFAULT_LICENSE_FILE,
        env_var_name='OTP_DEFAULT_LICENSE_FILE',
        allowed_types=str,
    )

    default_fault_tolerance = OtpProperty(
        description='Default value for USE_FT query property.',
        base_default='FALSE',
        env_var_name='OTP_DEFAULT_FAULT_TOLERANCE',
        allowed_types=str,
    )

    default_auth_username = OtpProperty(
        description='Default username used for authentication.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_DEFAULT_AUTH_USERNAME',
    )

    default_password = OtpProperty(
        description='Default password used for authentication.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_DEFAULT_PASSWORD',
    )

    http_address = OtpProperty(
        description='Default HTTP server used as WebAPI endpoint.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_HTTP_ADDRESS',
    )

    http_username = OtpProperty(
        description='Username used for WebAPI authentication.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_HTTP_USERNAME',
    )

    http_password = OtpProperty(
        description='Password used for WebAPI authentication.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_HTTP_PASSWORD',
    )

    http_proxy = OtpProperty(
        description='HTTP proxy used for WebAPI requests.',
        base_default=None,
        allowed_types=str,
        env_var_name='HTTP_PROXY',
    )

    https_proxy = OtpProperty(
        description='HTTPS proxy used for WebAPI requests.',
        base_default=None,
        allowed_types=str,
        env_var_name='HTTPS_PROXY',
    )

    access_token = OtpProperty(
        description='SSO access token for WebAPI endpoint.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_ACCESS_TOKEN',
    )

    client_id = OtpProperty(
        description='Client ID for obtaining SSO access token.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_CLIENT_ID',
    )

    client_secret = OtpProperty(
        description='Client Secret for obtaining SSO access token.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_CLIENT_SECRET',
    )

    access_token_url = OtpProperty(
        description='URL for obtaining SSO access token.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_ACCESS_TOKEN_URL',
    )

    trusted_certificates_file = OtpProperty(
        description='Path to the file with list of trusted Certificate Authority certificates for WebAPI requests.',
        base_default=None,
        allowed_types=str,
        env_var_name='OTP_SSL_CERT_FILE',
    )

    max_expected_ticks_per_symbol = OtpProperty(
        description='Expected maximum number of ticks per symbol (used for performance optimizations).',
        base_default=2000,
        allowed_types=int,
        env_var_name='OTP_MAX_EXPECTED_TICKS_PER_SYMBOL',
    )

    show_stack_info = OtpShowStackInfoProperty(
        description='Show stack info (filename and line or stack trace) in OneTick exceptions.',
        base_default=False,
        allowed_types=(str, bool, int),
        env_var_name='OTP_SHOW_STACK_INFO',
        env_var_func=OtpShowStackInfoProperty.parser,
    )

    log_symbol = OtpProperty(
        description='Log currently executed symbol. Note, this only works with unbound symbols. '
                    'Note, in this case :py:func:`otp.run<onetick.py.run>` does not produce the output '
                    'so it should be used only for debugging purposes.',
        base_default=False,
        allowed_types=(str, bool, int),
        env_var_name='OTP_LOG_SYMBOL',
        env_var_func=OtpShowStackInfoProperty.parser,
    )

    ignore_ticks_in_unentitled_time_range = OtpProperty(
        description='Default value for IGNORE_TICKS_IN_UNENTITLED_TIME_RANGE query property.',
        base_default=False,
        env_var_name='OTP_IGNORE_TICKS_IN_UNENTITLED_TIME_RANGE',
        allowed_types=(str, bool, int),
        env_var_func=OtpShowStackInfoProperty.parser,
    )

    main_query_generated_filename = OtpProperty(
        description='The name of the .otq file with generated main query executed by otp.run.',
        base_default='',
        env_var_name='OTP_MAIN_QUERY_GENERATED_FILENAME',
        allowed_types=str,
    )

    logging = OtpProperty(
        description='The logging level string or path to the file with configuration. '
                    'Check the documentation of python logging module for the configuration formats. '
                    'JSON format (in the file with .json suffix) and python configparser formats are supported.',
        base_default='WARNING',
        env_var_name='OTP_LOGGING',
        allowed_types=str,
    )

    otq_debug_mode = OtpProperty(
        description='Enable .otq files debug mode. '
                    'If set to True, onetick.py will keep all generated otq files and '
                    'log their paths to the console.',
        base_default=False,
        env_var_name='OTP_OTQ_DEBUG_MODE',
        allowed_types=(str, bool, int),
        env_var_func=OtpShowStackInfoProperty.parser,
    )

    allow_lowercase_in_saved_fields = OtpProperty(
        description='Allow using lower case characters in field names that are being stored in Onetick databases. '
                    'If set to False, onetick.py would not allow saving fields with lower case characters '
                    'to a database.',
        base_default=True,
        allowed_types=bool,
    )

    clean_up_tmp_files = OtpProperty(
        description='Control deleting temporary files created by onetick-py. '
                    'Temporary files are OneTick configuration files and generated .otq queries.',
        base_default=True,
        env_var_name='OTP_CLEAN_UP_TMP_FILES',
        allowed_types=(str, bool, int),
        env_var_func=OtpShowStackInfoProperty.parser,
    )

    default_schema_policy = OtpProperty(
        description='Default schema policy when querying onetick database. '
                    'See parameter ``schema_policy`` in :class:`otp.DataSource <onetick.py.DataSource>` '
                    'for the list of supported values.',
        base_default=None,
        env_var_name='OTP_DEFAULT_SCHEMA_POLICY',
        allowed_types=str,
    )

    disable_compatibility_checks = OtpProperty(
        description='Disable compatibility checks when querying OneTick database. '
                    'Using this parameter outside test environment could lead to unexpected errors.',
        base_default=False,
        allowed_types=bool,
        env_var_name='OTP_DISABLE_COMPATIBILITY_CHECKS',
    )


def get_options_table(cls):
    options_table = ('\n'
                     '.. csv-table::\n'
                     '   :header: "Name", "Environment Variable", "Description"\n'
                     '   :widths: auto\n\n')
    for name, prop in cls.get_changeable_config_options().items():
        name = f':py:attr:`otp.config.{name}<onetick.py.configuration.Config.{name}>`'
        options_table += f'   {name},"``{prop._env_var_name}``","{prop._description}"\n'
    for name, prop in cls.get_derived_config_options().items():
        name = f':py:attr:`otp.config.{name}<onetick.py.configuration.Config.{name}>`'
        options_table += f'   {name},"","{prop._description}"\n'
    return options_table


class OptionsTable:
    __doc__ = get_options_table(Config)


config = Config()



import os
import re

from locator_parser.common import Entity
from locator_parser.io import FileReader, PrintWriter
from locator_parser.actions import DoNothing


class Include(Entity):
    TAG = "INCLUDE"
    SINGLE = True
    RECURSIVELY = False

    def parse(self, reader, writer, action):
        path = self.path
        matched = re.search(r"\$\{(.*)\}", path)

        # env variable found in the path
        if matched:
            if matched.group(0)[2:-1] not in os.environ:
                raise RuntimeError("Includes use not defined environment variable: % s" % matched.group(0))

            path = path.replace(matched.group(0), os.environ[matched.group(0)[2:-1]])

        # read locator recursivly
        if Include.RECURSIVELY:
            parse_locator(FileReader(path), PrintWriter(), action, recursively=True)


class Includes(Entity):
    TAG = "INCLUDES"
    CHILDREN = [Include]


class Location(Entity):
    TAG = "LOCATION"
    SINGLE = True


class Locations(Entity):
    TAG = "LOCATIONS"
    CHILDREN = [Location]


class RawDB(Entity):
    TAG = "RAW_DB"
    CHILDREN = [Locations]
    HAS_PROPERTIES = True


class RawData(Entity):
    TAG = "RAW_DATA"
    CHILDREN = [RawDB]


class FeedOptions(Entity):
    TAG = "OPTIONS"
    SINGLE = True


class Feed(Entity):
    TAG = "FEED"
    HAS_PROPERTIES = True
    CHILDREN = [FeedOptions]


class DB(Entity):
    TAG = "DB"
    CHILDREN = [Locations, RawData, Feed]
    HAS_PROPERTIES = True


class DBs(Entity):
    TAG = "DATABASES"
    CHILDREN = [DB]


class ServerLocation(Entity):
    TAG = "LOCATION"
    SINGLE = True


class TickServers(Entity):
    TAG = "TICK_SERVERS"
    CHILDREN = [ServerLocation]


class CEPServerLocation(Entity):
    TAG = "LOCATION"
    SINGLE = True


class CEPTickServers(Entity):
    TAG = "CEP_TICK_SERVERS"
    CHILDREN = [CEPServerLocation]


class Range(Entity):
    TAG = "RANGE"
    SINGLE = True


class TimeInterval(Entity):
    TAG = "TIME_INTERVAL"
    HAS_PROPERTIES = True
    CHILDREN = [Range]


class VirtualDB(Entity):
    TAG = "DB"
    HAS_PROPERTIES = True
    CHILDREN = [TimeInterval]


class VirtualDBs(Entity):
    TAG = "VIRTUAL_DATABASES"
    CHILDREN = [VirtualDB]


def parse_locator(reader, writer, action=DoNothing(), recursively=False):
    Include.RECURSIVELY = recursively
    writer.refresh()
    reader.set_writer(writer)

    databases_p = DBs()
    v_databases_p = VirtualDBs()
    tick_servers_p = TickServers()
    cep_tick_servers_p = CEPTickServers()
    includes_p = Includes()

    for _ in reader:
        databases_p(reader, writer, action)
        v_databases_p(reader, writer, action)
        tick_servers_p(reader, writer, action)
        cep_tick_servers_p(reader, writer, action)
        includes_p(reader, writer, action)


# ----------------------------------------- #
# set parent
for _, cls in list(globals().items()):
    try:
        if issubclass(cls, Entity) and cls != Entity:
            for child in cls.CHILDREN:
                child.PARENT = Entity

                if not cls.HAS_PROPERTIES and not cls.SINGLE:
                    child.PARENT = cls
    except Exception:
        continue



import os
import re
from typing import Tuple, Optional, TYPE_CHECKING
from inspect import Parameter, formatannotation, Signature
from functools import wraps

from onetick.py.docs.docstring_parser import Docstring

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def param_doc(name,
              str_annotation=None,
              str_default=None,
              desc=None,
              annotation=Parameter.empty,
              default=Parameter.empty,
              kind=Parameter.POSITIONAL_OR_KEYWORD) -> Tuple[str, Parameter]:
    doc = f"""{name}"""

    if str_annotation is None and annotation is not Parameter.empty:
        str_annotation = formatannotation(annotation)

    if str_annotation:
        doc += f": {str_annotation}"
        if str_default:
            doc += f", default={str_default}"
        elif default is not Parameter.empty:
            doc += f", default={default}"
    if desc:
        doc += f"    {desc}"
    param = Parameter(name=name,
                      kind=kind,
                      default=default,
                      annotation=annotation)
    return doc, param


def docstring(parameters: Optional[list] = None, add_self=False):
    parameters = parameters or []

    def _decorator(fun):
        @wraps(fun)
        def _inner(*args, **kwargs):
            return fun(*args, **kwargs)
        doc = fun.__doc__ or ''
        doc = Docstring(doc)
        sig = []
        if add_self:
            sig.append(Parameter(name='self', kind=Parameter.POSITIONAL_OR_KEYWORD))
        sig.extend(doc.get_sig())
        for d, param in parameters:
            doc['Parameters'] = d
            sig.append(param)
        _inner.__doc__ = doc.build()
        _inner.__signature__ = Signature(sig)
        return _inner
    return _decorator


def alias(aliased_fun, doc_replacer=None, skip_ot_directives=True):
    """
    Returns new function with docstring and
    signature copied from ``aliased_fun``.
    """

    @wraps(aliased_fun)
    def fun(*args, **kwargs):
        return aliased_fun(*args, **kwargs)

    if doc_replacer:
        fun.__doc__ = doc_replacer(fun.__doc__)
    if skip_ot_directives:
        fun.__doc__ = re.sub('# OTdirective: .*', '', fun.__doc__)
    return fun


def is_windows():
    # PY-866: needed to be used in :skipif: sphinx doctest directives
    return os.name == 'nt'



from typing import TYPE_CHECKING, Collection, Optional, Union

from onetick.py.core.column import _Column
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


@inplace_operation
def sort(
    self: 'Source', by: Union[str, Collection[Union[str, '_Column']]], ascending=True, inplace=False
) -> Optional['Source']:
    """Sort ticks by columns.

    Parameters
    ----------
    by: str, Column or list of them
        Column(s) to sort by. It is possible to pass a list of column, where is the order is important:
        from the left to the right.
    ascending: bool or list
        Order to sort by. If list of columns is specified, then list of ascending values per column is expected.
        (the :class:`nan` is the smallest for ``float`` type fields)
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If ``inplace=True``, then it returns nothing. Otherwise method
        returns a new modified object

    Returns
    -------
    :class:`Source`

    See also
    --------
    **ORDER_BY** OneTick event processor

    Examples
    --------

    Single column examples

    >>> data = otp.Ticks({'X':[     94,   5,   34],
    ...                   'Y':[otp.nan, 3.1, -0.3]})
    >>> data = data.sort(data['X'])
    >>> otp.run(data)
                         Time   X    Y
    0 2003-12-01 00:00:00.001   5  3.1
    1 2003-12-01 00:00:00.002  34 -0.3
    2 2003-12-01 00:00:00.000  94  NaN

    >>> data = otp.Ticks({'X':[     94,   5,   34],
    ...                   'Y':[otp.nan, 3.1, -0.3]})
    >>> data = data.sort(data['Y'])
    >>> otp.run(data)
                         Time   X    Y
    0 2003-12-01 00:00:00.000  94  NaN
    1 2003-12-01 00:00:00.002  34 -0.3
    2 2003-12-01 00:00:00.001   5  3.1

    Inplace

    >>> data = otp.Ticks({'X':[     94,   5,   34],
    ...                   'Y':[otp.nan, 3.1, -0.3]})
    >>> data.sort(data['Y'], inplace=True)  # OTdirective: snippet-name: Arrange.sort.inplace;
    >>> otp.run(data)
                         Time   X    Y
    0 2003-12-01 00:00:00.000  94 NaN
    1 2003-12-01 00:00:00.002  34 -0.3
    2 2003-12-01 00:00:00.001  5   3.1

    Multiple columns

    >>> data = otp.Ticks({'X':[  5,   6,   3,   6],
    ...                   'Y':[1.4, 3.1, 9.1, 5.5]})
    >>> data = data.sort([data['X'], data['Y']])
    >>> otp.run(data)
                         Time  X    Y
    0 2003-12-01 00:00:00.002  3  9.1
    1 2003-12-01 00:00:00.000  5  1.4
    2 2003-12-01 00:00:00.001  6  3.1
    3 2003-12-01 00:00:00.003  6  5.5

    Ascending/descending control

    >>> data = otp.Ticks({'X':[     94,   5,   34],
    ...                   'Y':[otp.nan, 3.1, -0.3]})
    >>> data = data.sort(data['X'], ascending=False)
    >>> otp.run(data)
                         Time   X    Y
    0 2003-12-01 00:00:00.000  94  NaN
    1 2003-12-01 00:00:00.002  34 -0.3
    2 2003-12-01 00:00:00.001   5  3.1

    >>> # OTdirective: snippet-name: Arrange.sort.sort;
    >>> data = otp.Ticks({'X':[  5,   6,   3,   6],
    ...                   'Y':[1.4, 3.1, 9.1, 5.5]})
    >>> data = data.sort([data['X'], data['Y']], ascending=[True, False])
    >>> otp.run(data)
                         Time  X    Y
    0 2003-12-01 00:00:00.002  3  9.1
    1 2003-12-01 00:00:00.000  5  1.4
    2 2003-12-01 00:00:00.003  6  5.5
    3 2003-12-01 00:00:00.001  6  3.1
    """
    columns = by

    if isinstance(columns, list):
        objs = columns
    else:
        objs = [columns]

    if isinstance(ascending, list):
        asc_objs = ascending
    else:
        asc_objs = [ascending]

    items = []

    # -------------------------------
    # Columns processing
    # -------------------------------
    # convert to strings
    # TODO: it seems as a common code, need to move to a separate function
    for obj in objs:
        if isinstance(obj, _Column):
            items.append(obj.name)
        elif isinstance(obj, str):
            items.append(obj)
        else:
            # TODO: cover with tests
            raise TypeError(f"It is not supported to order by '{obj}' of type '{type(obj)}'")

    # validate
    for item in items:
        if item in self.__dict__:
            if not isinstance(self.__dict__[item], _Column):
                # TODO: cover with tests
                raise AttributeError(f"There is no '{item}' column")
            #  if
        else:
            # TODO: covert with tests
            raise AttributeError(f"There is no '{item}' column")

    # -------------------------------
    # Asc processing
    # -------------------------------
    asc_items = [True] * len(items)

    def asc_convert(v):
        return "ASC" if v else "DESC"

    for inx in range(len(items)):
        if inx >= len(asc_objs):
            asc_obj = asc_items[inx]
        else:
            asc_obj = asc_objs[inx]

        if isinstance(asc_obj, bool):
            asc_items[inx] = asc_convert(asc_obj)
        elif isinstance(asc_obj, int):
            asc_items[inx] = asc_convert(asc_obj)
        else:
            raise TypeError(f"asc can not be '{asc_obj}' of type '{type(asc_obj)}'")

    # ---------------
    # combine together
    order_by = [column_name + " " + asc for column_name, asc in zip(items, asc_items)]

    self.sink(otq.OrderByEp(order_by=",".join(order_by)))
    return self


def sort_values(self: 'Source', *args, **kwargs):
    """
    alias of sort

    See Also
    --------
    :meth:`Source.sort`
    """
    return self.sort(*args, **kwargs)



import datetime as dt
import inspect
import sys
import warnings
import math

from typing import Optional, Union, Type, Sequence

import onetick.py as otp
from onetick.py.otq import otq
import pandas as pd

import onetick.py.core._source
import onetick.py.functions
import onetick.py.db._inspection
from onetick.py.core.column import _Column
from onetick.py.core.source import Source

from .. import types as ott
from .. import utils, configuration
from ..core.column_operations._methods.methods import is_arithmetical
from ..core.column_operations.base import _Operation
from ..compatibility import is_supported_bucket_units_for_tick_generator
from onetick.py.aggregations._base import get_bucket_interval_from_datepart

from ..aggregations._docs import _bucket_time_doc
from onetick.py.docs.utils import docstring

from .common import get_start_end_by_date, update_node_tick_type, AdaptiveTickType
from .empty import Empty


def get_case_expr_with_datetime_limited_by_end_time(dt_expr: str) -> str:
    return f'CASE({dt_expr} > _END_TIME, 1, _END_TIME, {dt_expr})'


class Tick(Source):

    @docstring(parameters=[_bucket_time_doc], add_self=True)
    def __init__(
        self,
        data: Optional[dict] = None,
        offset=0,
        offset_part='millisecond',
        time: Optional[ott.datetime] = None,
        timezone_for_time=None,
        symbol=utils.adaptive_to_default,
        db=utils.adaptive_to_default,
        start=utils.adaptive,
        end=utils.adaptive,
        date=None,
        tick_type: Optional[AdaptiveTickType] = utils.adaptive,
        bucket_time: str = "start",
        bucket_interval: int = 0,
        bucket_units: Union[str, Type[utils.adaptive]] = utils.adaptive,
        num_ticks_per_timestamp: int = 1,
        **kwargs,
    ):
        """
        Generates a single tick for each bucket.
        By default a single tick for the whole query time interval is generated.

        Parameters
        ----------
        data: dict
            dictionary of columns names with their values.
            If specified, then parameter ``kwargs`` can't be used.
        offset: int, :ref:`datetime offset <datetime_offsets>`,\
                :py:class:`otp.timedelta <onetick.py.timedelta>`, default=0
            tick timestamp offset from query start time in `offset_part`
        offset_part: one of [nanosecond, millisecond, second, minute, hour, day, dayofyear, weekday, week, month, quarter, year], default=millisecond   #noqa
            unit of time to calculate ``offset`` from.
            Could be omitted if :ref:`datetime offset <datetime_offsets>` or
            :py:class:`otp.timedelta <onetick.py.timedelta>` objects are set as ``offset``.
        time: :py:class:`otp.datetime <onetick.py.datetime>`
            fixed time to set to all ticks.
            Note that this time should be inside time interval set by ``start`` and ``end`` parameters
            or by query time range.
        timezone_for_time: str
            timezone of the ``time``
        symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
            Symbol(s) from which data should be taken.
        db: str
            Database to use for tick generation
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            start time for tick generation. By default the start time of the query will be used.
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            end time for tick generation. By default the end time of the query will be used.
        date: :py:class:`otp.datetime <inetick.py.datetime>` – allows to specify a whole day
            instead of passing explicitly start and end parameters. If it is set along with
            the start and end parameters then last two are ignored.
        tick_type: str
            By default, the tick type value is not significant, and a placeholder string constant will be utilized.
            If you prefer to use the sink node's tick type instead of specifying your own,
            you can set the value to None.
        bucket_interval: int or :ref:`datetime offset objects <datetime_offsets>`
            Determines the length of each bucket (units depends on ``bucket_units``)
            for which the tick will be generated.

            Bucket interval can also be set via :ref:`datetime offset objects <datetime_offsets>`
            like :py:func:`otp.Second <onetick.py.Second>`, :py:func:`otp.Minute <onetick.py.Minute>`,
            :py:func:`otp.Hour <onetick.py.Hour>`, :py:func:`otp.Day <onetick.py.Day>`,
            :py:func:`otp.Month <onetick.py.Month>`.
            In this case you could omit setting ``bucket_units`` parameter.
        bucket_units: 'seconds', 'days' or 'months'
            Unit for value in ``bucket_interval``.
            Default is 'seconds'.
        num_ticks_per_timestamp: int
            The number of ticks to generate for every value of timestamp.
        kwargs:
            dictionary of columns names with their values.
            If specified, then parameter ``data`` can't be used.

        See also
        --------
        | **TICK_GENERATOR** OneTick event processor
        | :py:class:`otp.Ticks <onetick.py.Ticks>`

        Examples
        --------

        Simple usage, generate single tick:

        >>> t = otp.Tick(A=1, B='string', C=3.14, D=otp.dt(2000, 1, 1, 1, 1, 1, 1))
        >>> otp.run(t)
                Time  A       B     C                          D
        0 2003-12-01  1  string  3.14 2000-01-01 01:01:01.000001

        Generate single tick with offset:

        >>> t = otp.Tick(A=1, offset=otp.Minute(10))
        >>> otp.run(t)
                Time           A
        0 2003-12-01 00:10:00  1

        Generate one tick for each day in a week:

        >>> t = otp.Tick(A=1, start=otp.dt(2023, 1, 1), end=otp.dt(2023, 1, 8), bucket_interval=24 * 60 * 60)
        >>> otp.run(t)
                Time  A
        0 2023-01-01  1
        1 2023-01-02  1
        2 2023-01-03  1
        3 2023-01-04  1
        4 2023-01-05  1
        5 2023-01-06  1
        6 2023-01-07  1

        Generate tick every hour and add 1 minute offset to ticks' timestamps:

        >>> t = otp.Tick(A=1, offset=1, offset_part='minute', bucket_interval=60 * 60)
        >>> t.head(5)
                          Time  A
        0  2003-12-01 00:01:00  1
        1  2003-12-01 01:01:00  1
        2  2003-12-01 02:01:00  1
        3  2003-12-01 03:01:00  1
        4  2003-12-01 04:01:00  1

        Generate tick every hour and set fixed time:

        >>> t = otp.Tick(A=1, time=otp.dt(2023, 1, 2, 3, 4, 5, 6), bucket_interval=60 * 60,
        ...              start=otp.dt(2023, 1, 1), end=otp.dt(2023, 1, 8))
        >>> t.head(5)
                                Time  A
        0 2023-01-02 03:04:05.000006  1
        1 2023-01-02 03:04:05.000006  1
        2 2023-01-02 03:04:05.000006  1
        3 2023-01-02 03:04:05.000006  1
        4 2023-01-02 03:04:05.000006  1

        Use :ref:`datetime offset object <datetime_offsets>` as a ``bucket_interval``:

        .. testcode::
           :skipif: not is_supported_bucket_units_for_tick_generator()

           t = otp.Tick(A=1, bucket_interval=otp.Day(1))
           df = otp.run(t, start=otp.dt(2023, 1, 1), end=otp.dt(2023, 1, 5))
           print(df)

        .. testoutput::

                   Time  A
           0 2023-01-01  1
           1 2023-01-02  1
           2 2023-01-03  1
           3 2023-01-04  1
        """

        if self._try_default_constructor(**kwargs):
            return

        if data and kwargs:
            raise ValueError("Parameters 'data' and **kwargs can't be used at the same time")

        if data:
            kwargs = data

        if len(kwargs) == 0:
            raise ValueError("It is not allowed to have a tick without fields")

        if isinstance(offset, ott.OTPBaseTimeOffset):
            offset, offset_part = offset.get_offset()

            if offset < 0:
                raise ValueError("Negative offset not allowed")

            if offset_part not in {
                "nanosecond", "millisecond", "second", "minute", "hour", "day", "week", "month", "quarter", "year"
            }:
                raise ValueError(f"Unsupported DatePart passed to offset: {offset_part}")
        elif isinstance(offset, ott.timedelta):
            offset, offset_part = offset._get_offset()

        if time is not None and offset != 0:
            raise ValueError("It's not allowed to set parameter 'time' and set non-zero offset at the same time")

        bucket_time = self._get_bucket_time(bucket_time)
        if bucket_time == "BUCKET_END" and offset != 0:
            raise ValueError(
                "It's not allowed to set parameter 'bucket_time' to 'end' and set non-zero offset at the same time"
            )

        if isinstance(bucket_interval, ott.OTPBaseTimeOffset):
            bucket_interval, bucket_units = get_bucket_interval_from_datepart(bucket_interval)

        if date:
            # TODO: write a warning in that case
            start, end = get_start_end_by_date(date)

        columns = {}
        for key, value in kwargs.items():
            # the way to skip a field
            if value is None:
                continue

            if inspect.isclass(value):
                raise TypeError(f"Tick constructor expects values but not types, {value}")
            else:
                value_type = ott.get_object_type(value)

            if value_type is str:
                if isinstance(value, _Operation) or is_arithmetical(value):
                    if value.dtype is not str:
                        value_type = value.dtype
                elif len(value) > ott.string.DEFAULT_LENGTH:
                    value_type = ott.string[len(value)]

            if value_type is bool:
                value_type = float

            if issubclass(value_type, (ott.datetime, ott.date, dt.datetime, dt.date, pd.Timestamp)):
                value_type = ott.nsectime

            columns[key] = value_type

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(db=db,
                                               tick_type=tick_type,
                                               offset=offset,
                                               offset_part=offset_part,
                                               time=time,
                                               timezone_for_time=timezone_for_time,
                                               columns=columns,
                                               bucket_time=bucket_time,
                                               bucket_interval=bucket_interval,
                                               bucket_units=bucket_units,
                                               num_ticks_per_timestamp=num_ticks_per_timestamp,
                                               **kwargs),
            schema=columns,
        )

    def base_ep(self,
                db=utils.adaptive_to_default,
                tick_type=utils.adaptive,
                offset=0,
                offset_part='millisecond',
                time=None,
                timezone_for_time=None,
                columns=None,
                bucket_time="start",
                bucket_interval=0,
                bucket_units=utils.adaptive,
                num_ticks_per_timestamp=1,
                **kwargs):
        if columns is None:
            columns = {}

        params = ",".join(
            ott.type2str(columns[key]) + " " + str(key) + "=" + ott.value2str(value)
            for key, value in kwargs.items()
            if value is not None
        )

        tick_generator_kwargs = {}
        if bucket_units is not utils.adaptive:
            if is_supported_bucket_units_for_tick_generator(throw_warning=True):
                tick_generator_kwargs['bucket_interval_units'] = bucket_units.upper()
            elif bucket_units != 'seconds':
                raise ValueError("Parameter 'bucket_units' in otp.Tick is not supported on this OneTick version")

        src = Source(
            otq.TickGenerator(
                bucket_interval=bucket_interval,
                bucket_time=bucket_time,
                fields=params,
                num_ticks_per_timestamp=num_ticks_per_timestamp,
                **tick_generator_kwargs,
            ),
            schema=columns,
        )

        update_node_tick_type(src, tick_type, db)

        # TIMESTAMP += offset will add redundant nodes to sort the timestamps.
        # No sorting needed for a single tick.

        if offset:
            src.sink(
                otq.UpdateField(
                    field="TIMESTAMP",
                    value=get_case_expr_with_datetime_limited_by_end_time(
                        f"dateadd('{offset_part}', {offset}, TIMESTAMP, _TIMEZONE)"
                    )
                )
            )
        elif time:
            src.sink(otq.UpdateField(field="TIMESTAMP",
                                     value=ott.datetime2expr(time, timezone_naive=timezone_for_time)))
        return src

    @staticmethod
    def _get_bucket_time(bucket_time):
        if bucket_time == "BUCKET_START":
            warnings.warn("BUCKET_START value is deprecated. Please, use 'start' instead", FutureWarning)
        elif bucket_time == "BUCKET_END":
            warnings.warn("BUCKET_END value is deprecated. Please, use 'end' instead", FutureWarning)
        elif bucket_time == "start":
            bucket_time = "BUCKET_START"
        elif bucket_time == "end":
            bucket_time = "BUCKET_END"
        else:
            raise ValueError(f"Only 'start' and 'end' values supported as bucket time, but you've passed {bucket_time}")
        return bucket_time


def Ticks(data=None,  # NOSONAR
          symbol=utils.adaptive_to_default,
          db=utils.adaptive_to_default,
          start=utils.adaptive,
          end=utils.adaptive,
          tick_type: Optional[AdaptiveTickType] = utils.adaptive,
          timezone_for_time=None,
          offset=utils.adaptive,
          **inplace_data):
    """
    Data source that generates ticks.

    By default ticks are placed with the 1 millisecond offset from
    each other starting from the start of the query interval.

    The offset for each tick can be changed using the
    special reserved field name ``offset``, that specifies the time offset from the query start time.
    ``offset`` can be an integer, :ref:`datetime offset <datetime_offsets>` object
    or :py:class:`otp.timedelta <onetick.py.timedelta>`.

    Parameters
    ----------
    data: dict, list or pandas.DataFrame, optional
        Ticks values

        * ``dict`` -- <field_name>: <values>

        * ``list`` -- [[<field_names>], [<first_tick_values>], ..., [<n_tick_values>]]

        * :pandas:`DataFrame <pandas.DataFrame>`

        * ``None`` -- ``inplace_data`` will be used

    symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
        Symbol(s) from which data should be taken.
    db: str
        Database to use for tick generation
    start, end: :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>`, \
                    :py:class:`onetick.py.adaptive`
        Timestamp for data generation
    tick_type: str
        tick type for data generation
    timezone_for_time: str
        timezone for data generation
    offset: int, :ref:`datetime offset <datetime_offsets>` or :py:class:`otp.timedelta <onetick.py.timedelta>` \
            or list of such values or None
        Specifies the time offset for each tick from the query start time.
        Should be specified as the list of values, one for each tick,
        or as a single value that will be the same for all ticks.
        Special value None will disable changing timestamps for each tick,
        so all timestamps will be set to the query start time.
        Can't be used at the same time with the column `offset`.
    **inplace_data: list
        <field_name>: list(<field_values>)

    See also
    --------
    | **TICK_GENERATOR** OneTick event processor
    | **CSV_FILE_LISTING** OneTick event processor
    | :py:class:`otp.Tick <onetick.py.Tick>`

    Examples
    --------

    Pass the data as a dictionary:

    >>> d = otp.Ticks({'A': [1, 2, 3], 'B': [4, 5, 6]})
    >>> otp.run(d)
                         Time  A  B
    0 2003-12-01 00:00:00.000  1  4
    1 2003-12-01 00:00:00.001  2  5
    2 2003-12-01 00:00:00.002  3  6

    Pass the data using key-value arguments:

    >>> d = otp.Ticks(A=[1, 2, 3], B=[4, 5, 6])
    >>> otp.run(d)
                         Time  A  B
    0 2003-12-01 00:00:00.000  1  4
    1 2003-12-01 00:00:00.001  2  5
    2 2003-12-01 00:00:00.002  3  6

    Pass the data using list:

    >>> d = otp.Ticks([['A', 'B'],
    ...                [1, 4],
    ...                [2, 5],
    ...                [3, 6]])
    >>> otp.run(d)
                         Time  A  B
    0 2003-12-01 00:00:00.000  1  4
    1 2003-12-01 00:00:00.001  2  5
    2 2003-12-01 00:00:00.002  3  6

    Pass the data using :pandas:`pandas.DataFrame`.
    DataFrame should have a ``Time`` column containing datetime objects.

    >>> start_datetime = datetime.datetime(2023, 1, 1, 12)
    >>> time_array = [start_datetime + otp.Hour(1) + otp.Nano(1)]
    >>> a_array = [start_datetime - otp.Day(15) - otp.Nano(7)]
    >>> df = pd.DataFrame({'Time': time_array,'A': a_array})
    >>> data = otp.Ticks(df)
    >>> otp.run(data, start=start_datetime, end=start_datetime + otp.Day(1))
                               Time                             A
    0 2023-01-01 13:00:00.000000001 2022-12-17 11:59:59.999999993

    Example with setting ``offset`` for each tick:

    >>> data = otp.Ticks(X=[1, 2, 3], offset=[0, otp.Nano(1), 1])
    >>> otp.run(data)
                               Time  X
    0 2003-12-01 00:00:00.000000000  1
    1 2003-12-01 00:00:00.000000001  2
    2 2003-12-01 00:00:00.001000000  3

    Remove the ``offset`` for all ticks, in this case the timestamp of each tick is set to the start time of the query:

    >>> data = otp.Ticks(X=[1, 2, 3], offset=None)
    >>> otp.run(data)
            Time  X
    0 2003-12-01  1
    1 2003-12-01  2
    2 2003-12-01  3

    Parameter ``offset`` allows to set the same value for all ticks:

    >>> data = otp.Ticks(X=[1, 2, 3], offset=otp.Nano(13))
    >>> otp.run(data)
                               Time  X
    0 2003-12-01 00:00:00.000000013  1
    1 2003-12-01 00:00:00.000000013  2
    2 2003-12-01 00:00:00.000000013  3
    """

    if db is utils.adaptive_to_default:
        db = configuration.config.get('default_db')

    if isinstance(data, pd.DataFrame):
        if offset is not utils.adaptive:
            raise ValueError("Parameter 'offset' can't be set when passing pandas.DataFrame.")
        if data.empty:
            warnings.warn('otp.Ticks got empty DataFrame as input, returning otp.Empty', stacklevel=2)
            return Empty(db=db, symbol=symbol, tick_type=tick_type, start=start, end=end)
        if 'Time' not in data.columns:
            raise ValueError('Field `Time` is required for constructing an `otp.Source` from `pandas.DataFrame`')
        data = data.rename(columns={"Time": "time"})
        # to_dict('list') doesn't work correctly with pandas timestamps on some versions
        data = data.to_dict('series')
        data = {column_name: series.to_list() for column_name, series in data.items()}

    if data and len(inplace_data) != 0:
        raise ValueError("Data can be passed only using either the `data` parameter "
                         "or inplace through the key-value args")

    if isinstance(data, list):
        reform = {}
        for inx, key in enumerate(data[0]):
            reform[key] = [sub_list[inx] for sub_list in data[1:]]

        data = reform

    if data is None:
        if inplace_data:
            data = inplace_data
        else:
            raise ValueError("You don't specify any date to create ticks from. "
                             "Please, use otp.Empty for creating empty data source")
    else:
        data = data.copy()

    def check_value_len(_data):
        # check all columns to have the same number of rows
        value_len = None
        for key, value in _data.items():
            if value_len is None:
                value_len = len(value)
            elif value_len != len(value):
                raise ValueError(
                    f"It is not allowed to have different columns of different lengths, "
                    f"some of columns have {value_len} length, but column '{key}', as instance, has {len(value)}"
                )
        return value_len

    value_len = check_value_len(data)

    use_absolute_time = False

    if 'offset' in data:
        if offset is not utils.adaptive:
            raise ValueError("Parameter 'offset' and column 'offset' can't be set at the same time.")
        else:
            offset = data['offset']

    if offset is not utils.adaptive:
        if "time" in data:
            raise ValueError("You cannot specify 'offset' column/parameter and 'time' column at the same time.")
    elif "time" in data:
        use_absolute_time = True
    else:
        # by default the difference between each offset is 1 (1 millisecond)
        offset = list(range(value_len))

    disable_offsets = False
    if offset is None:
        disable_offsets = True
    elif offset is not utils.adaptive and not isinstance(offset, Sequence):
        # if offset is set as a single value then just copy-paste it for all rows
        offset = [offset] * value_len

    if offset is not utils.adaptive and offset is not None:
        data['offset'] = offset
        check_value_len(data)

    if not use_absolute_time and not disable_offsets:
        offset_values = []
        offset_parts = []
        for ofv in data['offset']:
            if isinstance(ofv, ott.offsets.Tick):
                offset_values.append(ofv.n)
                try:
                    str_repr = str(ofv.datepart)[1:-1]
                except Exception:
                    str_repr = str(ofv.base)[1:-1].lower()
                offset_parts.append(str_repr)
            elif isinstance(ofv, ott.timedelta):
                value, str_repr = ofv._get_offset()
                offset_values.append(value)
                offset_parts.append(str_repr)
            else:
                offset_values.append(ofv)
                offset_parts.append('millisecond')
        data['offset'] = offset_values
        data['offset_part'] = offset_parts

    def split_data_for_tick(columns):
        tick_parameters = {}
        tick_columns = {}
        for key, value in columns.items():
            if key in {'offset', 'offset_part', 'time'}:
                tick_parameters[key] = value
            else:
                tick_columns[key] = value
        return tick_columns, tick_parameters

    if value_len == 1:
        columns = {key: value[0] for key, value in data.items()}
        tick_columns, tick_parameters = split_data_for_tick(columns)
        return Tick(tick_columns, db=db, symbol=symbol, tick_type=tick_type, start=start, end=end,
                    timezone_for_time=timezone_for_time, **tick_parameters)
    else:
        # select only columns that do not contain None there to support
        # heterogeneous data
        not_none_columns = []
        for key in data.keys():
            data[key] = [float(elem) if isinstance(elem, bool) else elem for elem in data[key]]
        for key, value in data.items():
            add = True
            for v in value:
                # we need it, because can't use _Column instances in if-clauses
                if isinstance(v, _Column):
                    continue
                if v is None:
                    add = False
                    break

            if add:
                not_none_columns.append(key)

        # if a field is a onetick operation, it cannot be csv'd (it's dynamic)
        is_outside_data_dependent = False
        for key, value in data.items():
            for v in value:
                if isinstance(v, _Operation):
                    is_outside_data_dependent = True
                    break

        # infinity() and (on windows) nan() cannot be natively read from a csv
        has_special_values = False
        for key, value in data.items():
            for v in value:
                if isinstance(v, ott._inf) or \
                    (isinstance(v, ott._nan) or isinstance(v, float) and math.isnan(v)) \
                        and sys.platform.startswith("win"):
                    has_special_values = True
                    break

        if (len(not_none_columns) == len(data)) and (not is_outside_data_dependent) and (not has_special_values):
            # Data is homogenous; CSV backing can be used
            return _DataCSV(data, value_len, db=db, symbol=symbol, tick_type=tick_type, start=start, end=end,
                            timezone_for_time=timezone_for_time, use_absolute_time=use_absolute_time,
                            disable_offsets=disable_offsets)
        else:
            # Fallback is a merge of individual ticks
            ticks = []

            for inx in range(value_len):
                columns = {key: value[inx] for key, value in data.items()}
                tick_columns, tick_parameters = split_data_for_tick(columns)
                ticks.append(Tick(tick_columns, db=db, symbol=symbol, tick_type=tick_type, start=start, end=end,
                                  timezone_for_time=timezone_for_time, **tick_parameters))

            return onetick.py.functions.merge(ticks, align_schema=not_none_columns)


class _DataCSV(Source):
    def __init__(
        self,
        data=None,
        length=None,
        db=utils.adaptive_to_default,
        symbol=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        use_absolute_time=False,
        timezone_for_time=None,
        disable_offsets=False,
        **kwargs,
    ):
        if self._try_default_constructor(**kwargs):
            return

        if data is None or length is None:
            raise ValueError("'data' and 'length' parameters can't be None")

        def datetime_to_expr(v):
            if ott.is_time_type(v):
                return ott.datetime2expr(v, timezone_naive=timezone_for_time)
            if isinstance(v, ott.nsectime):
                # TODO: change to ott.value2str after PY-441
                return f'NSECTIME({v})'
            if isinstance(v, ott.msectime):
                return ott.value2str(v)
            raise ValueError(f"Can't convert value {v} to datetime expression")

        if use_absolute_time:
            # converting values of "time" column to onetick expressions
            converted_times = []
            for d in data["time"]:
                converted_times.append(datetime_to_expr(d))
            data["time"] = converted_times

        def csv_rep(value):
            if issubclass(type(value), str):
                return '"' + value.replace("\\", "\\\\").replace('"', '\\"') + '"'
            else:
                return str(value)

        def get_type_of_column(key):
            def get_type_of_value(value):
                t = ott.get_object_type(value)

                if ott.is_time_type(t):
                    return ott.nsectime
                elif t is str:
                    if len(value) <= ott.string.DEFAULT_LENGTH:
                        return str
                    else:
                        return ott.string[len(value)]
                else:
                    return t

            types = [get_type_of_value(v) for v in data[key]]
            res, _ = utils.get_type_that_includes(types)
            return res

        columns = {key: get_type_of_column(key) for key in data}

        expression_columns = []
        header_columns = {}
        for key in list(columns):
            header_columns[key] = columns[key]
            # converting values of datetime columns to onetick expressions
            if columns[key] is ott.nsectime:
                data[key] = [datetime_to_expr(v) for v in data[key]]
                header_columns[key] = get_type_of_column(key)
                expression_columns.append(key)

        transposed_data = [[csv_rep(value[i]) for key, value in data.items()] for i in range(length)]

        text_header = ",".join(f"{ott.type2str(v)} {k}" for k, v in header_columns.items())
        text_data = "\n".join([",".join(data_row) for data_row in transposed_data])

        if use_absolute_time:
            del columns["time"]
        elif not disable_offsets:
            del columns["offset"]
            del columns["offset_part"]

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(columns=columns,
                                               db=db,
                                               tick_type=tick_type,
                                               use_absolute_time=use_absolute_time,
                                               text_header=text_header,
                                               text_data=text_data,
                                               expression_columns=expression_columns,
                                               disable_offsets=disable_offsets),
            schema=columns,
        )

    def base_ep(self, columns, db, tick_type, use_absolute_time, text_header, text_data, expression_columns=None,
                disable_offsets=False):

        node = Source(
            otq.CsvFileListing(
                discard_timestamp_column=True,
                time_assignment="_START_TIME",
                field_delimiters="','",
                quote_chars='"""',
                handle_escaped_chars=True,
                file_contents=text_data,
                first_line_is_title=False,
                fields=text_header,
            ),
            schema=columns,
        )

        update_node_tick_type(node, tick_type, db)

        if use_absolute_time:
            # don't trust UpdateField
            node.sink(otq.AddField(field='____TMP____', value="EVAL_EXPRESSION(time, 'datetime')"))
            node.sink(otq.UpdateField(field="TIMESTAMP", value="____TMP____"))
            node.sink(otq.Passthrough(fields="time,____TMP____", drop_fields="True"))
            node.sink(otq.OrderBy(order_by="TIMESTAMP ASC"))
        elif not disable_offsets:
            node.sink(
                otq.AddField(
                    field='nsectime ____TMP____',
                    value=get_case_expr_with_datetime_limited_by_end_time(
                        "dateadd(offset_part, offset, TIMESTAMP, _TIMEZONE)"
                    )
                )
            )
            node.sink(otq.OrderBy(order_by="____TMP____ ASC"))
            node.sink(otq.UpdateField(field="TIMESTAMP", value="____TMP____"))
            node.sink(otq.Passthrough(fields="offset,offset_part,____TMP____", drop_fields="True"))
            node.sink(otq.OrderBy(order_by="TIMESTAMP ASC"))

        for column in expression_columns or []:
            # don't trust UpdateField
            node.sink(otq.RenameFields(f'{column}=____TMP____'))
            node.sink(otq.AddField(field=column, value="EVAL_EXPRESSION(____TMP____, 'datetime')"))
            node.sink(otq.Passthrough(fields='____TMP____', drop_fields=True))
        node.sink(otq.Table(keep_input_fields=True,
                            fields=', '.join(f'nsectime {column}' for column in expression_columns)))

        return node


def TTicks(data):  # NOSONAR
    """
    .. deprecated:: 1.3.101

    Transposed Ticks format.

    Parameters
    ----------
    data: list
        list of list, where the first sublist is the header, and other are values
    """

    warnings.warn("The nice and helpful function `TTicks` is going to be deprecated. "
                  "You could use the `Ticks` to pass data in the same format there",
                  FutureWarning)

    data_dict = {}

    for inx, key in enumerate(data[0]):
        data_dict[key] = [sub_list[inx] for sub_list in data[1:]]

    return Ticks(data_dict)



import datetime
import sys
import warnings

from typing import Optional

import pytz
import dateutil.tz
import tzlocal

import onetick.py as otp
from onetick.py.backports import zoneinfo


def get_tzfile_by_name(timezone):
    if isinstance(timezone, str):
        try:
            timezone = pytz.timezone(timezone)
        except pytz.exceptions.UnknownTimeZoneError:
            timezone = dateutil.tz.gettz(timezone)
    return timezone


def get_local_timezone_name():
    tz = tzlocal.get_localzone()
    try:
        return tz.zone  # type: ignore
    except AttributeError:
        return tz.key  # type: ignore


def get_timezone_from_datetime(dt) -> Optional[str]:
    tzinfo = getattr(dt, 'tzinfo', None)
    if tzinfo is None:
        return None
    if tzinfo is datetime.timezone.utc:
        return 'UTC'
    if isinstance(tzinfo, pytz.BaseTzInfo):
        return tzinfo.zone
    if isinstance(tzinfo, zoneinfo.ZoneInfo):
        return tzinfo.key
    if isinstance(tzinfo, dateutil.tz.tzlocal):
        return get_local_timezone_name()
    if isinstance(tzinfo, dateutil.tz.tzstr) and hasattr(tzinfo, '_s'):
        return tzinfo._s
    if isinstance(tzinfo, dateutil.tz.tzfile):
        if sys.platform == 'win32':
            warnings.warn(
                "It's not recommended to use dateutil.tz timezones on Windows platform. "
                "Function 'get_timezone_from_datetime' can't guarantee correct results in this case. "
                "Please, use pytz or zoneinfo timezones instead."
            )
        if hasattr(tzinfo, '_filename'):
            if tzinfo._filename == '/etc/localtime':
                return get_local_timezone_name()
            for timezone in zoneinfo.available_timezones():
                if tzinfo._filename.endswith(timezone):
                    return timezone
    if sys.platform == 'win32':
        if isinstance(tzinfo, dateutil.tz.win.tzwin) and hasattr(tzinfo, '_name'):
            return tzinfo._name
    raise ValueError(f"Can't get timezone name from datetime '{dt}' with tzinfo {tzinfo}")


def convert_timezone(dt, src_timezone, dest_timezone) -> datetime.datetime:
    """
    Converting timezone-naive ``dt`` object localized in ``src_timezone`` timezone
    to the specified ``dest_timezone`` and returning also timezone-naive object.
    """
    if src_timezone is None:
        src_timezone = get_local_timezone_name()
    # using pandas, because stdlib datetime has some bug around epoch on Windows
    dt = otp.datetime(dt).ts
    # change timezone-naive to timezone-aware
    dt = dt.tz_localize(src_timezone)
    # convert timezone
    dt = dt.tz_convert(dest_timezone)
    # change timezone-aware to timezone-naive
    dt = dt.tz_localize(None)
    # convert to datetime
    dt = dt.to_pydatetime()
    return dt



import warnings
from typing import TYPE_CHECKING, List, Optional, Union

from onetick import py as otp
from onetick.py.otq import otq
from onetick.py.functions import __copy_sources_on_merge_or_join
from onetick.py.aggregations._docs import (
    _boundary_tick_bucket_doc,
    _bucket_end_condition_doc,
    _bucket_interval_doc,
    _bucket_time_doc,
    _bucket_units_doc,
    _end_condition_per_group_doc,
    _group_by_doc,
    _groups_to_display_doc,
)
from onetick.py.docs.utils import docstring, param_doc
from onetick.py.aggregations._base import _Aggregation


if TYPE_CHECKING:
    from onetick.py.core.source import Source


def __add__(self: 'Source', other: 'Source') -> 'Source':
    return otp.merge([self, other])


def append(self: 'Source', other) -> 'Source':
    """
    Merge data source with `other`

    Parameters
    ----------
    other: List, Source
        data source to merge

    Returns
    -------
    Source
    """
    if isinstance(other, list):
        return otp.merge(other + [self])
    else:
        return otp.merge([self, other])


def diff(self: 'Source', other: 'Source',
         fields: Optional[Union[str, List[str]]] = None,
         ignore: bool = False,
         output_ignored_fields: Optional[bool] = None,
         show_only_fields_that_differ: Optional[bool] = None,
         show_matching_ticks: Optional[bool] = None,
         non_decreasing_value_fields: Optional[Union[str, List[str]]] = None,
         threshold: Optional[int] = None,
         left_prefix: str = 'L',
         right_prefix: str = 'R',
         drop_index: bool = True) -> 'Source':
    """
    Compare two time-series.

    A tick from the first time series is considered to match a tick from the second time series
    if both ticks have identical ``non_decreasing_value_fields`` values
    and matching values of all ``fields`` that are present in both ticks and are not listed as the fields to be ignored.

    A field is considered to match another field
    if both fields have identical names, comparable types, and identical values.

    Field names in an output tick are represented as <source_name>.<field_name>.

    Note
    ----
    The first source is considered to be the base for starting the matching process.
    If a match is found, all ticks in the second source that occur before the first matched tick
    are considered different from those in the first source.
    The subsequent searches for the matches in the second source will begin after this latest matched tick.
    The ticks from the second source that are between the found matches
    are considered different from the ticks in the first source.

    Parameters
    ----------
    fields:
        List of fields to be used or ignored while comparing input time series.
        By default, if this value is not set, all fields are used in comparison.
    ignore:
        If True, fields specified in the ``fields`` parameter are ignored while comparing the input time series.
        Otherwise, by default, only the fields that are specified in the ``fields`` parameter are used in comparison.
        If the ``fields`` parameter is empty, the value of this parameter is ignored.
    output_ignored_fields:
        If False, the fields which are not used in comparison are excluded from the output.
        Otherwise, by default, all fields are propagated.
    show_only_fields_that_differ:
        If True (default), the method outputs only the tick fields the values of which were different.
        But if ``output_ignored_fields`` is set to True, then ignored fields are still propagated.
    show_matching_ticks:
        If True, the output of this method consists of matched ticks from both input time series
        instead of unmatched ticks.
        The output tick timestamp is equal to the earliest timestamp of its corresponding input ticks.
        Default value is False.
    non_decreasing_value_fields:
        List of *non-decreasing* value fields to be used for matching.
        If value of this parameter is **TIMESTAMP** (default), it compares two time series based on tick timestamp.
        If other field is specified, a field named <source_name>.<TIMESTAMP>
        will be added to the tick whose value equals the tick's primary timestamp.
    threshold:
        Specifies the number of first diff ticks to propagate.
        By default all such ticks are propagated.
    left_prefix:
        The prefix used in the output for fields from the left source.
    right_prefix:
        The prefix used in the output for fields from the right source.
    drop_index: bool
        If False, the output tick will also carry the position(s) of input tick(s)
        in input time series in field <source_name>.INDEX for each source.
        The lowest position number is 1.
        If True then <source_name>.INDEX fields will not be included in the output.

    Returns
    -------
    :class:`Source`

    See Also
    --------
    **DIFF** OneTick event processor

    Examples
    --------

    Print all ticks that have any unmatched fields:

    >>> t = otp.Ticks(A=[1, 2], B=[0, 0])
    >>> q = otp.Ticks(A=[1, 3], B=[0, 0])
    >>> data = t.diff(q)
    >>> otp.run(data)
                         Time  L.A  R.A
    0 2003-12-01 00:00:00.001    2    3

    Also show fields that were not different:

    >>> t = otp.Ticks(A=[1, 2], B=[0, 0])
    >>> q = otp.Ticks(A=[1, 3], B=[0, 0])
    >>> data = t.diff(q, show_only_fields_that_differ=False)
    >>> otp.run(data)
                         Time  L.A  L.B  R.A  R.B
    0 2003-12-01 00:00:00.001    2    0    3    0

    Change prefixes for output fields:

    >>> t = otp.Ticks(A=[1, 2], B=[0, 0])
    >>> q = otp.Ticks(A=[1, 3], B=[0, 0])
    >>> data = t.diff(q, left_prefix='LEFT', right_prefix='RIGHT')
    >>> otp.run(data)
                         Time  LEFT.A  RIGHT.A
    0 2003-12-01 00:00:00.001       2        3

    If there are several matching ticks then only the first will be matched:

    .. testcode::
       :skipif: not otp.compatibility.is_diff_show_matching_ticks_supported()

       t = otp.Ticks(A=[1, 1, 1, 1, 1], B=[1, 2, 3, 4, 5], offset=[0, 0, 1000, 2000, 2000])
       q = otp.Ticks(A=[1, 1, 1, 1, 1], B=[3, 4, 5, 6, 7], offset=[0, 1000, 1000, 2000, 2000])
       data = t.diff(q, fields=['A'], show_matching_ticks=True, output_ignored_fields=True)
       print(otp.run(data))

    .. testoutput::

                        Time  L.A  L.B  R.A  R.B
       0 2003-12-01 00:00:00    1    1    1    3
       1 2003-12-01 00:00:01    1    3    1    4
       2 2003-12-01 00:00:02    1    4    1    6
       3 2003-12-01 00:00:02    1    5    1    7
    """

    if not fields:
        fields = []
    if isinstance(fields, str):
        fields = [fields]
    fields = list(map(str, fields))
    for field in fields:
        if field not in self.schema or field not in other.schema:
            raise ValueError(f"Field {field} is not in schema")

    if threshold is None:
        threshold = ''  # type: ignore
    elif threshold < 0:
        raise ValueError("Parameter 'threshold' must be non-negative")

    ep_params = dict(
        fields=','.join(map(str, fields)),
        ignore=ignore,
        threshold=threshold,
    )

    if non_decreasing_value_fields is not None:
        if not non_decreasing_value_fields:
            raise ValueError("Parameter 'non_decreasing_value_fields' can't be empty")
        if isinstance(non_decreasing_value_fields, str):
            non_decreasing_value_fields = [non_decreasing_value_fields]
        non_decreasing_value_fields = list(map(str, non_decreasing_value_fields))
        for field in non_decreasing_value_fields:
            if field not in self.schema or field not in other.schema:
                raise ValueError(f"Field {field} is not in schema")
        if otp.compatibility.is_diff_non_decreasing_value_fields_supported():
            ep_params['non_decreasing_value_fields'] = ','.join(map(str, non_decreasing_value_fields))
        else:
            warnings.warn("Parameter 'non_decreasing_value_fields' is not supported on this version of OneTick")

    if show_only_fields_that_differ and output_ignored_fields:
        raise ValueError(
            "Parameters 'output_ignored_fields' and 'show_only_fields_that_differ' can't be set at the same time"
        )
    if show_only_fields_that_differ is None and output_ignored_fields is None:
        if ignore:
            ep_params['output_ignored_fields'] = True
        else:
            ep_params['output_ignored_fields'] = False
        ep_params['show_only_fields_that_differ'] = not ep_params['output_ignored_fields']
    elif show_only_fields_that_differ is None:
        ep_params['show_only_fields_that_differ'] = not output_ignored_fields
    elif output_ignored_fields is None:
        ep_params['output_ignored_fields'] = False
        ep_params['show_only_fields_that_differ'] = show_only_fields_that_differ

    if show_matching_ticks is not None:
        if otp.compatibility.is_diff_show_matching_ticks_supported():
            ep_params['show_matching_ticks'] = show_matching_ticks
        else:
            warnings.warn("Parameter 'show_matching_ticks' is not supported on this version of OneTick")

        if ep_params.get('show_matching_ticks') and show_only_fields_that_differ is None:
            ep_params['show_only_fields_that_differ'] = False

    schema = {
        f'{left_prefix}.INDEX': int,
        f'{right_prefix}.INDEX': int,
    }
    for src_prefix, src_schema in [(left_prefix, self.schema), (right_prefix, other.schema)]:
        for field, dtype in src_schema.items():
            if ignore and field in fields and not output_ignored_fields:
                continue
            schema[f'{src_prefix}.{field}'] = dtype

    result = otp.Source(
        node=otq.Diff(**ep_params),
        schema=schema,
    )
    __copy_sources_on_merge_or_join(result, (self, other),
                                    names=(left_prefix, right_prefix))
    if drop_index:
        result = result.drop([f'{left_prefix}.INDEX', f'{right_prefix}.INDEX'])

    return result


def lee_and_ready(self: 'Source', qte: 'Source',
                  quote_delay: float = 0.0,
                  show_quote_fields: bool = False) -> 'Source':
    """
    Adds a numeric attribute to each tick in the stream of trade ticks,
    the value of which classifies the trade as a buy, a sell, or undefined.

    This is an implementation of the Lee and Ready algorithm:
    Match up a trade with the most recent good quote that is at least X seconds older than the trade —

    * if the trade's price is closer to the ask price, label trade a buy (1);
    * else, if it is closer to the bid price, label it a sell (-1);
    * else, if trade's price is at the mid-quote, then if it is higher than the last trade's price,
      classify it as a buy (1);
    * else, if it is less, classify it as a sell (-1);
    * else, if it is the same, classify it the same way as the previous trade was classified.
    * If all of these fail, classify the trade as unknown (0).

    This method expects two sources as its input: source of trades (``self``) and source of quotes (``qte``).
    While ticks propagated by trades source should have the PRICE,SIZE fields,
    ticks propagated by ``qte`` source should have the ASK_PRICE,ASK_SIZE,BID_PRICE,BID_SIZE fields.

    Output of this method is a time series of trades ticks
    with the Lee and Ready indicator field (**BuySellFlag**) added.

    Parameters
    ----------
    qte:
        The source of quotes.
    quote_delay:
        The minimum number of seconds that needs to elapse between the trade and the quote
        before the quote can be considered for a join with the trade.

        The value is a float number.
        Only the first three digits of the fraction are currently used,
        thus the highest supported granularity of quote delay is milliseconds.
        Sub-millisecond parts of the trade's and the quote's timestamps are ignored when computing delay between them.
    show_quote_fields:
        If set to True, the quote fields that classified trade will also be shown for each trade.
        Note that if there were no quotes before trade, then quote fields will be set to 0.

    Returns
    -------
    :class:`Source`

    See Also
    --------
    **LEE_AND_READY** OneTick event processor

    Examples
    --------

    Add field **BuySellFlag** to the ``trd`` source:

    >>> import os
    >>> trd = otp.CSV(os.path.join(csv_path, 'trd.csv'))
    >>> qte = otp.CSV(os.path.join(csv_path, 'qte.csv'))
    >>> data = trd.lee_and_ready(qte)
    >>> otp.run(data).head(5)
                            Time   PRICE  SIZE  BuySellFlag
    0 2003-12-01 09:00:00.086545  178.26   246         -1.0
    1 2003-12-01 09:00:00.245208  178.26     1          1.0
    2 2003-12-01 09:00:00.245503  178.26     1          1.0
    3 2003-12-01 09:00:00.387100  178.21     9          1.0
    4 2003-12-01 09:00:00.387105  178.21    12          1.0

    Fields from ``qte`` can be added with ``show_quote_fields`` parameter:

    >>> data = trd.lee_and_ready(qte, show_quote_fields=True)
    >>> data = data.drop(['ASK_SIZE', 'BID_SIZE'])
    >>> otp.run(data).head(5)
                            Time   PRICE  SIZE  BuySellFlag               QTE_TIMESTAMP  ASK_PRICE  BID_PRICE
    0 2003-12-01 09:00:00.086545  178.26   246         -1.0  2003-12-01 09:00:00.028307     178.80     177.92
    1 2003-12-01 09:00:00.245208  178.26     1          1.0  2003-12-01 09:00:00.244626     178.57     177.75
    2 2003-12-01 09:00:00.245503  178.26     1          1.0  2003-12-01 09:00:00.244626     178.57     177.75
    3 2003-12-01 09:00:00.387100  178.21     9          1.0  2003-12-01 09:00:00.387096     178.57     177.75
    4 2003-12-01 09:00:00.387105  178.21    12          1.0  2003-12-01 09:00:00.387096     178.57     177.75

    Set ``quote_delay`` parameter to 300 milliseconds:

    >>> data = trd.lee_and_ready(qte, show_quote_fields=True, quote_delay=0.3)
    >>> data = data.drop(['ASK_SIZE', 'BID_SIZE'])
    >>> otp.run(data).head(5)
                            Time   PRICE  SIZE  BuySellFlag               QTE_TIMESTAMP  ASK_PRICE  BID_PRICE
    0 2003-12-01 09:00:00.086545  178.26   246          0.0  1969-12-31 19:00:00.000000        0.0       0.00
    1 2003-12-01 09:00:00.245208  178.26     1          0.0  1969-12-31 19:00:00.000000        0.0       0.00
    2 2003-12-01 09:00:00.245503  178.26     1          0.0  1969-12-31 19:00:00.000000        0.0       0.00
    3 2003-12-01 09:00:00.387100  178.21     9         -1.0  2003-12-01 09:00:00.087540      180.0     177.62
    4 2003-12-01 09:00:00.387105  178.21    12         -1.0  2003-12-01 09:00:00.087540      180.0     177.62
    """

    schema = self.schema.copy()
    if show_quote_fields:
        schema.update(**qte.schema, **{'QTE_TIMESTAMP': otp.nsectime})
    schema.update(**{'BuySellFlag': float})

    result = otp.Source(
        node=otq.LeeAndReady(
            quote_delay=quote_delay,
            show_quote_fields=show_quote_fields
        ),
        schema=schema,
    )
    __copy_sources_on_merge_or_join(result, (self, qte), names=('TRD', 'QTE'))

    return result


_smallest_time_granularity_msec_name_doc = param_doc(
    name='smallest_time_granularity_msec',
    annotation=int,
    desc="""
    This method works by first sampling the source tick series with a constant rate.
    This is the sampling interval (1 / rate).
    As a consequence, any computed delay will be divisible by this value.
    It is important to carefully choose this parameter, as this method has a computational cost of O(N * log(N))
    per bucket, where N = (*duration_of_bucket_in_msec* + ``max_ts_delay_msec``) / ``smallest_time_granularity_msec``.
    Default: 1.
    """,
    default=1,
)


_max_ts_delay_msec_doc = param_doc(
    name='max_ts_delay_msec',
    annotation=int,
    desc="""
     The known upper bound on the delay's magnitude.
    The computed delay will never be greater than this value.
    Default: 1000.
    """,
    default=1000,
)


@docstring(
    parameters=[
        _smallest_time_granularity_msec_name_doc,
        _max_ts_delay_msec_doc,
        _bucket_interval_doc,
        _bucket_time_doc,
        _bucket_units_doc,
        _bucket_end_condition_doc,
        _end_condition_per_group_doc,
        _boundary_tick_bucket_doc,
        _group_by_doc,
        _groups_to_display_doc,
    ],
    add_self=True,
)
def estimate_ts_delay(self: 'Source', other: 'Source',
                      input_field1_name: str, input_field2_name: str,
                      **kwargs) -> 'Source':
    """
    Given two time series of ticks, computes how much delay the second series has in relation to the first.
    A negative delay should be interpreted as the first series being delayed instead.

    The two series do not necessarily have to be identical with respect to the delay,
    nor do they have to represent the same quantity (i.e. be of the same magnitude).
    The only requirement to get meaningful results is that the two series be (linearly) correlated.

    Output ticks always have 2 fields:

        * *DELAY_MS*, which is the computed delay in milliseconds, and
        * *CORRELATION*, which is the Zero-Normalized Cross-Correlation of the two time series
          after that delay is applied.

    Parameters
    ----------
    other: Source
        The other source.
    input_field1_name: str
        The name of the compared field from the first source.
    input_field2_name: str
        The name of the compared field from the second source.

    Returns
    -------
    :class:`Source`

    See Also
    --------
    **ESTIMATE_TS_DELAY** OneTick event processor

    Examples
    --------

    Calculating delay between the same sources will result in DELAY_MSEC equal to 0.0 and CORRELATION equal to 1.0:
    (Note that correlation method may return NaN values for smaller buckets):

    .. testcode::
       :skipif: not otp.compatibility.is_supported_estimate_ts_delay()

       import os
       trd = otp.CSV(os.path.join(csv_path, 'trd.csv'))
       other = trd.deepcopy()
       data = trd.estimate_ts_delay(other, 'PRICE', 'PRICE', bucket_interval=10, bucket_time='start')
       df = otp.run(data, start=otp.dt(2003, 12, 1, 9), end=otp.dt(2003, 12, 1, 10))
       print(df)

    .. testoutput::

                          Time  DELAY_MSEC  CORRELATION
       0   2003-12-01 09:00:00         0.0          1.0
       1   2003-12-01 09:00:10         0.0          1.0
       2   2003-12-01 09:00:20         0.0          1.0
       3   2003-12-01 09:00:30         0.0          1.0
       4   2003-12-01 09:00:40         0.0          1.0
       ..                  ...         ...          ...
       355 2003-12-01 09:59:10         0.0          1.0
       356 2003-12-01 09:59:20         0.0          1.0
       357 2003-12-01 09:59:30         NaN          NaN
       358 2003-12-01 09:59:40         0.0          1.0
       359 2003-12-01 09:59:50         0.0          1.0

       [360 rows x 3 columns]

    Try changing timestamps of other time-series to see how delay values are changed:

    .. testcode::
       :skipif: not otp.compatibility.is_supported_estimate_ts_delay()

       import os
       trd = otp.CSV(os.path.join(csv_path, 'trd.csv'))
       other = trd.deepcopy()
       other['TIMESTAMP'] += otp.Milli(5)
       data = trd.estimate_ts_delay(other, 'PRICE', 'PRICE', bucket_interval=10, bucket_time='start')
       df = otp.run(data, start=otp.dt(2003, 12, 1, 9), end=otp.dt(2003, 12, 1, 10))
       print(df)

    .. testoutput::

                          Time  DELAY_MSEC  CORRELATION
       0   2003-12-01 09:00:00        -5.0          1.0
       1   2003-12-01 09:00:10        -5.0          1.0
       2   2003-12-01 09:00:20        -5.0          1.0
       3   2003-12-01 09:00:30        -5.0          1.0
       4   2003-12-01 09:00:40        -5.0          1.0
       ..                  ...         ...          ...
       355 2003-12-01 09:59:10        -5.0          1.0
       356 2003-12-01 09:59:20        -5.0          1.0
       357 2003-12-01 09:59:30         NaN          NaN
       358 2003-12-01 09:59:40        -5.0          1.0
       359 2003-12-01 09:59:50        -5.0          1.0

       [360 rows x 3 columns]

    Try filtering out some ticks from other time-series to see how delay and correlation values are changed:

    .. testcode::
       :skipif: not otp.compatibility.is_supported_estimate_ts_delay()

       import os
       trd = otp.CSV(os.path.join(csv_path, 'trd.csv'))
       other = trd.deepcopy()
       other = other[::2]
       data = trd.estimate_ts_delay(other, 'PRICE', 'PRICE', bucket_interval=10, bucket_time='start')
       df = otp.run(data, start=otp.dt(2003, 12, 1, 9), end=otp.dt(2003, 12, 1, 10))
       print(df)

    .. testoutput::

                          Time  DELAY_MSEC  CORRELATION
       0   2003-12-01 09:00:00         0.0     1.000000
       1   2003-12-01 09:00:10         0.0     1.000000
       2   2003-12-01 09:00:20         0.0     1.000000
       3   2003-12-01 09:00:30         0.0     0.999115
       4   2003-12-01 09:00:40     -1000.0     0.706111
       ..                  ...         ...          ...
       355 2003-12-01 09:59:10         0.0     0.983786
       356 2003-12-01 09:59:20         0.0     1.000000
       357 2003-12-01 09:59:30         NaN          NaN
       358 2003-12-01 09:59:40      -306.0     0.680049
       359 2003-12-01 09:59:50         0.0     0.752731

       [360 rows x 3 columns]
    """
    if not otp.compatibility.is_supported_estimate_ts_delay():
        raise RuntimeError('estimate_ts_delay() is not supported on this OneTick version')

    if input_field1_name not in self.schema:
        raise ValueError(f"Field '{input_field1_name}' is not in the schema of the first source.")

    if input_field2_name not in other.schema:
        raise ValueError(f"Field '{input_field2_name}' is not in the schema of the second source.")

    schema = {'DELAY_MSEC': float, 'CORRELATION': float}

    smallest_time_granularity_msec = kwargs.pop('smallest_time_granularity_msec', 1)
    max_ts_delay_msec = kwargs.pop('max_ts_delay_msec', 1000)

    # we only use this class for validation of common parameters
    class EstimateTsDelay(_Aggregation):
        EP = otq.EstimateTsDelay
        NAME = 'ESTIMATE_TS_DELAY'
        FIELDS_TO_SKIP = ['column_name', 'running', 'all_fields', 'output_field_name']

    common_ep_params = {k.lower(): v for k, v in EstimateTsDelay('TIMESTAMP', **kwargs).ep_params.items()}

    result = otp.Source(
        node=otq.EstimateTsDelay(
            input_field1_name=input_field1_name,
            input_field2_name=input_field2_name,
            smallest_time_granularity_msec=smallest_time_granularity_msec,
            max_ts_delay_msec=max_ts_delay_msec,
            **common_ep_params,
        ),
        schema=schema,
    )
    __copy_sources_on_merge_or_join(result, (self, other))
    return result



from typing import Optional

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source
from onetick.py.core.column_operations.base import OnetickParameter

from .. import types as ott
from .. import utils

from .common import update_node_tick_type, AdaptiveTickType


class SymbologyMapping(Source):
    _PROPERTIES = Source._PROPERTIES + ["_p_dest_symbology"]

    def __init__(
        self,
        dest_symbology: Optional[str] = None,
        tick_type: Optional[AdaptiveTickType] = utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        symbols=utils.adaptive,
        schema=None,
        **kwargs,
    ):
        """
        Shows symbology mapping information for specified securities stored in the reference database.

        Input (source) symbology is taken from the input symbol,
        if it has a symbology part in it (e.g., RIC::REUTERS::MSFT),
        or defaults to that of the input database, which is specified in the locator file.

        Parameter ``symbol_date`` must be set in :py:func:`otp.run <onetick.py.run>`
        for this source to work.

        Parameters
        ----------
        dest_symbology: str, :py:class:`otp.param <onetick.py.core.column_operations.base.OnetickParameter>`
            Specifying the destination symbology for symbol translation.
        tick_type: str
            Tick type to set on the OneTick's graph node.
            Can be used to specify database name with tick type or tick type only.

            By default setting this parameter is not required, database is usually set
            with parameter ``symbols`` or in :py:func:`otp.run <onetick.py.run>`.
        start:
            Custom start time of the source.
            If set, will override the value specified in :py:func:`otp.run <onetick.py.run>`.
        end:
            Custom end time of the source.
            If set, will override the value specified in :py:func:`otp.run <onetick.py.run>`.
        symbols:
            Symbol(s) from which data should be taken.
            If set, will override the value specified in :py:func:`otp.run <onetick.py.run>`.

        Examples
        --------

        Getting mapping for OID symbology for one symbol:

        >>> data = otp.SymbologyMapping(dest_symbology='OID')
        >>> otp.run(data, symbols='US_COMP::AAPL',  # doctest: +SKIP
        ...         symbol_date=otp.dt(2022, 1, 3),
        ...         date=otp.dt(2022, 1, 3))
                Time END_DATETIME MAPPED_SYMBOL_NAME
        0 2022-01-03   2022-01-04               9706

        Getting mapping for all symbols in `US_COMP` database in single source:

        >>> data = otp.SymbologyMapping(dest_symbology='OID')
        >>> data = otp.merge([data],
        ...                  symbols=otp.Symbols('US_COMP', keep_db=True),
        ...                  identify_input_ts=True)
        >>> data = data[['SYMBOL_NAME', 'MAPPED_SYMBOL_NAME']]
        >>> otp.run(data,  # doctest: +SKIP
        ...         symbol_date=otp.dt(2022, 1, 3),
        ...         date=otp.dt(2022, 1, 3))
                    Time    SYMBOL_NAME MAPPED_SYMBOL_NAME
        0     2022-01-03     US_COMP::A               3751
        1     2022-01-03    US_COMP::AA             647321
        2     2022-01-03   US_COMP::AAA             695581
        3     2022-01-03  US_COMP::AAAU             673522
        4     2022-01-03   US_COMP::AAC             703090
        ...          ...            ...                ...
        11746 2022-01-03   US_COMP::ZWS             273584
        11747 2022-01-03    US_COMP::ZY             704054
        11748 2022-01-03  US_COMP::ZYME             655470
        11749 2022-01-03  US_COMP::ZYNE             633589
        11750 2022-01-03  US_COMP::ZYXI             208375
        """
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if dest_symbology is None:
            raise TypeError("Missing required argument: 'dest_symbology'")

        if isinstance(dest_symbology, OnetickParameter):
            dest_symbology = dest_symbology.parameter_expression

        self._p_dest_symbology = dest_symbology

        super().__init__(
            _symbols=symbols,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(dest_symbology, tick_type),
            schema=schema,
            **kwargs,
        )

        self.schema['MAPPED_SYMBOL_NAME'] = str
        self.schema['END_DATETIME'] = ott.nsectime

    @property
    def dest_symbology(self):
        return self._p_dest_symbology

    def base_ep(self, dest_symbology, tick_type):
        src = Source(otq.SymbologyMapping(dest_symbology=dest_symbology))
        update_node_tick_type(src, tick_type)
        return src



# A common place to put backports for specific versions of Python

import sys

if sys.version_info >= (3, 8):
    from typing import Literal
    from functools import singledispatchmethod, lru_cache
else:
    from typing_extensions import Literal
    from singledispatchmethod import singledispatchmethod
    from backports.functools_lru_cache import lru_cache


if sys.version_info >= (3, 9):
    import zoneinfo
    from functools import cached_property, cache
    import ast
    astunparse = ast.unparse
else:
    from backports.cached_property import cached_property
    from astunparse import unparse as astunparse
    from backports import zoneinfo

    def cache(user_function):
        'Simple lightweight unbounded cache.  Sometimes called "memoize".'
        return lru_cache(maxsize=None)(user_function)



import os
import abc
import subprocess
import pandas as pd

from datetime import datetime
import onetick.py as otp

from . import utils, configuration


class _LicenseBase(abc.ABC):
    def __init__(self, file, dir):
        self._dir = dir
        self._file = file

        if otp.__webapi__:
            return

        if self._file is None:
            raise ValueError("Something goes wrong, you pass None instead of path to license file")

        if self._dir is not None and not os.path.exists(self._dir):
            raise FileNotFoundError(f'"{self._dir}" license directory does not exist')
        if not os.path.exists(self._file):
            raise FileNotFoundError(f'"{self._file}" license file does not exist')

    @property
    def dir(self):
        return self._dir

    @property
    def file(self):
        return self._file


class Default(_LicenseBase):
    """
    Default license path
    """

    def __init__(self):
        default_file = configuration.config.default_license_file
        default_dir = configuration.config.default_license_dir
        if default_file is None and not otp.__webapi__:
            raise ValueError(
                "I can't understand the default license path :( Please, use "
                "onetick.py.license.Custom to specify a path to license."
            )

        super().__init__(default_file, default_dir)


class Remote(_LicenseBase):

    def __init__(self):
        # remote license doesn't use local files
        pass

    @property
    def dir(self):
        return None

    @property
    def file(self):
        return None


class Custom(_LicenseBase):
    """
    Custom license path
    """

    # it is not useless
    # pylint: disable=useless-parent-delegation
    def __init__(self, file, directory=None):
        """
        Parameters
        ----------
        file: str, os.PathLike
            Path to license file.
        directory: str, os.PathLike, optional
            Path to license directory. Default is None, that means it is not used.
        """
        super().__init__(file, directory)


class Server(_LicenseBase):
    """
    License that is based on license server. It generates new license file
    """

    def __init__(self, addresses=None, file=None, directory=None, reload=0):
        """
        Parameters
        ----------
        addresses: list of strings
            List of ip address or domain names with ports where a license server is running.
            For example: ['license.server.A:123', 'license.server.B:345']
        file: str or PathLike, optional
            License file path. Passed file path should exist. If file is not set or equal to None,
            then temporary file will be used (temporary file lives while an instance lives).
            NOTE: passed license file might be updated according to the 'reload' parameter.
        directory: str or PathLike, optional
            License directory path. Passed directory should exist. If directory is not set or
            equal to None, then temporary directory will be used.
        reload: 0 or offsets from onetick.py, ie onetick.py.Day, Hour, Month etc
            Indicator to reload license from license server. 0 means that reload on creation,
            offset means that if license file modification time more than offset, then it will
            be updated from the server.
        """
        self._generated = False

        if file is None:
            file = utils.TmpFile()
        else:
            if isinstance(reload, pd.tseries.offsets.Tick):
                modify = datetime.fromtimestamp(os.path.getmtime(file))
                current = datetime.now()
                diff_ms = (current - modify).total_seconds() * 1000000
                expire_ms = reload.nanos / 1000

                if diff_ms < expire_ms:
                    self._generated = True
            elif isinstance(reload, int):
                if reload == 0:
                    # always generate
                    self._generated = False
                else:
                    raise ValueError("You've passed not supported value to the 'reload' parameter")
            else:
                raise ValueError("You've passed not supported value to the 'reload' parameter")

        if directory is None:
            directory = utils.TmpDir()

        super().__init__(file, directory)

        self._servers = ",".join(addresses)

        if not self._generated:
            self.__generate()

    @property
    def servers(self):
        return self._servers

    def __generate(self):
        class EnvVarScope:
            def __init__(self, cfg):
                self._env = None
                self._cfg = cfg

            def __enter__(self):
                if "ONE_TICK_CONFIG" in os.environ:
                    self._env = os.environ["ONE_TICK_CONFIG"]
                os.environ["ONE_TICK_CONFIG"] = str(self._cfg)

            def __exit__(self, exc_type, exc_val, exc_tb):
                if self._env is None:
                    if "ONE_TICK_CONFIG" in os.environ:
                        del os.environ["ONE_TICK_CONFIG"]
                else:
                    os.environ["ONE_TICK_CONFIG"] = self._env

        tmp_locator = utils.tmp_locator(empty=True)
        tmp_config = utils.tmp_config(
            locator=tmp_locator, license_path=str(self.file), license_dir=self.dir, license_servers=self.servers
        )

        with EnvVarScope(tmp_config):
            self.__generate_license()

        self._generated = True

    def __generate_license(self):
        lic_generator = self.__license_getter_path(os.path.join(utils.omd_dist_path(), "one_tick", "bin"))

        subprocess.run(
            [lic_generator, "-features", "COLL,QUERY,LOAD"],
            stdout=subprocess.PIPE, stderr=subprocess.PIPE,
            check=True,
        )

    @staticmethod
    def __license_getter_path(omd_dist_path):
        license_getter = os.path.join(omd_dist_path, "local_license_getter.exe")
        assert os.path.exists(license_getter), f"license getter is not found by {license_getter}"

        return license_getter
